import numpy as np
import mec.data

def test_mec_lp_LP():
    from mec.lp import LP   
    data = mec.data.load_stigler_data(verbose = True)
    dietLP = LP(data['N_i_j'].T,data['c_j'],data['d_i'])
    dietLP.gurobi_solve(verbose=0)
    return

def test_mec_lp_Dictionary():
    from mec.lp import Dictionary
    dictionary_example = Dictionary(A_i_j = np.array([[2, 1], [1, 2]]),
                                d_i = np.array([2,2]),
                                c_j = np.array([1,1]),
                                slack_var_names_i = ['s_1', 's_2'],
                                decision_var_names_j = ['x_1', 'x_2']
                                )
    dictionary_example.display()
    path_example = [dictionary_example.primal_solution(verbose=1)]
    dictionary_example.plot2d(path_example)
    dictionary_example.simplex_loop(verbose=3)
    [N_i_j,minallowance_i,unitcost_j,_,_] = mec.data.load_stigler_data(7,5).values()
    stigler_dictionary = Dictionary(N_i_j.T, unitcost_j , minallowance_i,
                             ['s_' + str(j) for j in range(N_i_j.shape[1])],
                             ['π_' + str(i) for i in range(N_i_j.shape[0])])
    stigler_dictionary.simplex_loop(verbose=2)[1]
    return
    
def test_mec_lp_Tableau():
    from mec.lp import Tableau
    from mec.data import load_stigler_data
    [N_i_j,minallowance_i,unitcost_j,_,_] = load_stigler_data().values()
    stigler_tableau = Tableau(N_i_j.T, unitcost_j , minallowance_i,
                             ['s_' + str(j) for j in range(N_i_j.shape[1])],
                             ['π_' + str(i) for i in range(N_i_j.shape[0])])

    stigler_tableau.simplex_solve()[1]
    return
########################################################
########################################################
########################################################
def test_mec_ot_OTProblem(limited=False):
    from mec.ot import OTProblem
    [data_X,data_Y,A_k_l] =  mec.data.load_DupuyGalichon_data().values()
    sdX,sdY = data_X.std().values, data_Y.std().values
    mX,mY = data_X.mean().values, data_Y.mean().values
    feats_x_k, feats_y_l = ((data_X-mX)/sdX).values, ((data_Y-mY)/sdY).values
    nbx,nbk = feats_x_k.shape
    nby,nbl = feats_y_l.shape
    Φ_x_y = feats_x_k @ A_k_l @ feats_y_l.T
    if limited:
        Φ_x_y = Φ_x_y[0:10,0:10]
    marriageEx = OTProblem(Φ_x_y)
    (μ_x_y,u_x,v_y) = marriageEx.solve_full_lp()
    print('Man 0 matches with woman '+str(np.argwhere(μ_x_y[0,:] != 0)[0][0])+'.'  )
    (μ_x_y,u_x,v_y) = marriageEx.solve_partial_lp()
    print(μ_x_y.sum())
    u_x,v_y = marriageEx.solve_dual_partial_multiobj_lp(favor_Y=True)
    print(u_x.min(),v_y.min())
    if limited:
        nbx,nby = 10,8
    else:
        nbx,nby = 50,30
    marriage_ex = OTProblem(Φ_x_y[:nbx,:nby],np.ones(nbx) / nbx, np.ones(nby) / nby)
    nrow , ncol = min(8, nbx) , min(8, nby)
    marriage_ex.solve_full_lp()
    marriage_ex.logdomainIPFP_with_LSE_trick(0.5)
    marriage_ex.solveGLM(0.5 )
    marriage_ex.matrixIPFP(0.5 )

    return
    
def test_mec_ot_TULogit():
    def build_surplus(nbx,nby):
        xs, ys = np.repeat(range(1,nbx+1),nby).reshape(nbx,nby)/nbx, np.repeat(range(1,nby+1),nbx).reshape(nbx,nby).T/nby
        phi1_xy = -((xs-ys)**2).flatten()
        phimat = np.vstack([phi1_xy,
                            -((xs-ys)**2 * ((xs+ys)/2)**2).flatten(),
                             -( (xs-ys)**2 * ((xs+ys-2)/2)**2).flatten(),
                            -( (xs-ys)**2 *(xs+ys-1)**2).flatten()
                           ]).T
        phimat_mean = phimat.mean(axis = 0 , keepdims=True)
        phimat_stdev = phimat.std(axis = 0 , keepdims=True)
        φ_xy_k = np.hstack([np.ones((nbx*nby,1)) ,(phimat - phimat_mean)/phimat_stdev]  )
        _,nbk = φ_xy_k.shape
        return np.vstack( [φ_xy_k ,np.zeros( (nbx,nbk) ),np.zeros( (nby,nbk) )])

    from mec.ot import TUlogit
    [cs_μhat_a, cs_Nhat,cs_nbx,cs_nby] = mec.data.load_ChooSiow_data().values()
    choo_siow_mkt = TUlogit(cs_nbx,cs_nby,build_surplus(cs_nbx,cs_nby),cs_μhat_a)  
    uv_GS_diy,λ_GS_diy_k = choo_siow_mkt.fit_diy(10000,tol = 1e-8)
    uv_GS_GLM, λ_GS_GLM_k = choo_siow_mkt.fit_glm()
    print('λ_GS_GLM_k = '+str(λ_GS_GLM_k))
    print('λ_GS_diy_k = '+str(λ_GS_diy_k))
    print(choo_siow_mkt.assess(np.append(uv_GS_GLM, λ_GS_GLM_k)))
    print(choo_siow_mkt.assess(np.append(uv_GS_diy, λ_GS_diy_k)))
    choo_siow_mkt.isCoercive()
    return
########################################################
########################################################
def test_mec_gt_Matrix_game():
    from mec.gt import Matrix_game
    Phi_i_j = np.array([[50, 100, 93.8], 
                    [93.9, 60.0, 82.8], 
                    [97.6, 100, 73.2]])
    thegame = Matrix_game(Phi_i_j)
    print(thegame.BRI(1) ,thegame.BRJ(0) )
    thegame.compute_eq()
    thegame.minimax_LP()
    thegame.minimax_CP()



def test_mec_gt_Bimatrix_game():
    from mec.gt import Bimatrix_game
    penalty_data = np.array([[50, 100, 93.8], [93.9, 60.0, 82.8], [97.6, 100, 73.2]])

    penalty_zero_sum = Bimatrix_game(A_i_j = penalty_data, B_i_j = 100 - penalty_data)
    penalty_nonzero_sum = Bimatrix_game(A_i_j = penalty_data,
                                    B_i_j = np.array([[150, 100, 100], [100, 150, 100], [100, 100, 150]]) - penalty_data)
    penalty_zero_sum.mangasarian_stone_solve()
    penalty_nonzero_sum.lemke_howson_solve()

########################################################
########################################################
########################################################

def test_mec_lp():
    test_mec_lp_LP()
    test_mec_lp_Dictionary()
    test_mec_lp_Tableau()
    return


def test_mec_ot(limited=False):
    test_mec_ot_OTProblem(limited)
    test_mec_ot_TULogit()
    
    return
    
def test_mec_gt():
    test_mec_gt_Matrix_game()
    test_mec_gt_Bimatrix_game()
    
    return

def test_mec(limited=False):
    print('Linear programming tests:\n'+'*'*30)
    test_mec_lp()
    print('*'*30+'\nOptimal transport tests:\n'+'*'*30)
    test_mec_ot(limited)
    print('*'*30+'\nGame theory tests:\n'+'*'*30)
    test_mec_gt()
    print('*'*30+'\n'+'*'*30+'\nAll tests completed successfully.')
    return
    



import gurobipy as grb
import numpy as np
import sympy
from sympy import *
from mec.lp import Dictionary
from mec.lp import Tableau


class Matrix_game:
    def __init__(self, A_i_j):
        self.nbi,self.nbj = A_i_j.shape
        self.A_i_j = A_i_j

    def BRI(self,j):
        return np.argwhere(self.A_i_j[:,j] == np.max(self.A_i_j[:,j])).flatten()

    def BRJ(self,i):
        return np.argwhere(self.A_i_j[i,:] == np.min(self.A_i_j[i,:])).flatten()

    def purestrat_solve(self):
        return [ (i,j) for j in range(self.nbj) for i in range(self.nbi) if ( (i in self.BRI(j)) and (j in self.BRJ(i)) ) ]

    def solve(self, verbose=0):
        model = grb.Model()
        model.Params.OutputFlag = 0
        y = model.addMVar(shape=self.nbj)
        model.setObjective(np.ones(self.nbj) @ y, grb.GRB.MAXIMIZE)
        model.addConstr(self.A_i_j @ y <= np.ones(self.nbi))
        model.optimize()
        ystar = np.array(model.getAttr('x'))
        xstar = np.array(model.getAttr('pi'))
        V_2 = 1 / model.getAttr('ObjVal')
        p_i = V_2 * xstar
        q_j = V_2 * ystar
        if verbose > 0: print('p_i =', p_i, '\nq_j =', q_j)
        return {'p_i': p_i, 'q_j': q_j, 'val': V_2}

    def simplex_solve(self, verbose=0):
        tableau = Tableau(A_i_j = self.A_i_j, d_i = np.ones(self.nbi), c_j = np.ones(self.nbj),
                          decision_var_names_j = ['y_'+str(j) for j in range(self.nbj)])
        ystar, xstar, ystar_sum = tableau.simplex_solve()
        p_i = xstar / ystar_sum
        q_j = ystar / ystar_sum
        if verbose > 0: print('p_i =', p_i, '\nq_j =', q_j)
        return {'p_i': p_i, 'q_j': q_j, 'val': 1/ystar_sum}

    def chambolle_pock_solve(self, tol=10e-6, max_iter=10000):
        L1 = np.max(np.abs(self.A_i_j))
        sigma, tau = 1/L1, 1/L1

        p_i = np.ones(self.nbi) / self.nbi
        q_j = np.ones(self.nbi) / self.nbj
        q_prev = q_j.copy()

        gap = np.inf
        i=0
        while (gap >  tol) and (i < max_iter):
            q_tilde = 2*q_j - q_prev
            p_i *= np.exp(-sigma * self.A_i_j @ q_tilde)
            p_i /= p_i.sum()

            q_prev = q_j.copy()
            q_j *= np.exp(tau * self.A_i_j.T @ p_i)
            q_j /= q_j.sum()
            gap = np.max(self.A_i_j.T @ p_i) - np.min(self.A_i_j @ q_j)
            i += 1
        return p_i, q_j, gap, i


class LCP: # z >= 0, w = M z + q >= 0, z.w = 0
    def __init__(self, M_i_j, q_i, z_names_i = None, w_names_i = None):
        if M_i_j.shape[0] != M_i_j.shape[1]:
            raise ValueError("M_i_j must be square.")
        if M_i_j.shape[0] != len(q_i):
            raise ValueError("M_i_j and q_i must be of the same size.")
        if z_names_i = None :
            z_names_i = ['z_'+str(i+1) for i in range(self.nbi)]
        if w_names_i = None :
            w_names_i = ['w_'+str(i+1) for i in range(self.nbi)]
        self.M_i_j, self.q_i = M_i_j, q_i
        self.nbi = len(q_i)
        self.z_names_i, self.w_names_i = z_names_i, w_names_i

    def qp_solve(self, silent=True, verbose=0):
        qp = grb.Model()
        if silent:
            qp.Params.OutputFlag = 0
        qp.Params.NonConvex = 2
        z = qp.addMVar(shape = self.nbi)
        w = qp.addMVar(shape = self.nbi)
        qp.addConstr(w - self.M_i_j @ z == self.q_i)
        qp.setObjective(z @ w, sense = grb.GRB.MINIMIZE)
        qp.optimize()
        zsol, wsol, obj = z.x, w.x, qp.ObjVal
        print('z.w =', obj)
        if verbose > 0:
            print('z =', zsol)
            print('w =', wsol)
        return zsol, wsol, obj

    def plot_cones(self):
        if self.nbi != 2:
            raise ValueError('Can\'t plot in 2D because the problem is of order different from 2.')
        A, q = -self.M_i_j, self.q_i
        I = np.eye(2)
        fig = plt.figure(figsize=(5, 5))
        ax = fig.add_subplot(1, 1, 1)
        ax.spines['left'].set_position('zero'), ax.spines['bottom'].set_position('zero')
        ax.spines['right'].set_color('none'), ax.spines['top'].set_color('none')
        r = np.array([[.6,.7],[.8,.9]])
        angles = np.array([[0, np.pi/2], [None, None]])
        for j in range(2):
            if any(A[:,j] != 0): angles[1,j] = (-1)**(A[1,j]<0) * np.arccos(A[0,j]/np.linalg.norm(A[:,j]))
        for i in range(2):
            for j in range(2):
                angle_1 = angles[i,0]
                angle_2 = angles[j,1]
                if angle_1 != None and angle_2 != None:
                    if abs(angle_1 - angle_2) < np.pi:
                        theta = np.linspace(np.minimum(angle_1, angle_2), np.maximum(angle_1, angle_2), 100)
                    elif abs(angle_1 - angle_2) > np.pi:
                        theta = np.linspace(np.maximum(angle_1, angle_2), np.minimum(angle_1, angle_2) + 2*np.pi, 100)
                    ax.fill(np.append(0, r[i,j]*np.cos(theta)), np.append(0, r[i,j]*np.sin(theta)), 'b', alpha=0.1)
        for i in range(2):
            ax.quiver(0, 0, I[0,i], I[1,i], scale=1, scale_units='xy', color='b')
            ax.text(I[0,i], I[1,i], r'$E_{{{}}}$'.format(i+1), fontsize=12, ha='left', va='bottom')
            if any(A[:,i] != 0):
                ax.quiver(0, 0, A[0,i]/np.linalg.norm(A[:,i]), A[1,i]/np.linalg.norm(A[:,i]), scale=1, scale_units='xy', color='b')
                ax.text(A[0,i]/np.linalg.norm(A[:,i]), A[1,i]/np.linalg.norm(A[:,i]), r'$-M_{{\cdot{}}}$'.format(i+1), fontsize=12, ha='left', va='bottom')
        ax.quiver(0, 0, q[0]/np.linalg.norm(q), q[1]/np.linalg.norm(q), angles='xy', scale_units='xy', scale=1, color='r')
        plt.xlim(-1.2, 1.2), plt.ylim(-1.2, 1.2)
        plt.show()

    def create_tableau(self, display=False):
        if decision_var_names_j is None :
            decision_var_names_j = ['z_'+str(i+1) for i in range(self.nbi)]
        if slack_var_names_i is None :
            slack_var_names_i = ['w_'+str(i+1) for i in range(self.nbi)]
        tab = Tableau(A_i_j = -np.block([self.M_i_j, np.ones((self.nbi,1))]),
                      d_i = self.q_i, c_j = None,
                      decision_var_names_j=self.z_names_i+['z_0'], slack_var_names_i=self.w_names_i)
        self.tableau = tab
        if display: tab.display()

    def initialize_basis(self, verbose=0, display=False):
        self.create_tableau()
        zis = self.tableau.decision_var_names_j
        wis = self.tableau.slack_var_names_i
        kent = 2*self.nbi # z_0 enters
        kdep = np.argmin(self.q_i) # w_istar departs
        self.tableau.update(kent, kdep)
        if verbose>0: print((wis+zis)[kent] + ' enters, ' + (wis+zis)[kdep] + ' departs')
        if display: self.tableau.display()
        return kdep

    def lemke_solve(self, verbose=0, maxit=100):
        counter = 0
        if all(self.q_i >= 0):
            print('==========')
            print('Solution found: trivial LCP (q >= 0).')
            zsol, wsol = np.zeros(self.nbi), self.q_i
            return zsol, wsol
        kdep = self.initialize_basis(verbose=verbose-1)
        zis = self.tableau.decision_var_names_j
        wis = self.tableau.slack_var_names_i
        complements = list(self.nbi+np.arange(self.nbi)) + list(np.arange(self.nbi))
        while counter < maxit:
            counter += 1
            kent = complements[kdep]
            kdep = self.tableau.determine_departing(kent)
            if kdep is None:
                break
            if verbose > 1:
                print('Basis: ', [(wis+zis)[i] for i in self.tableau.k_b])
                print((wis+zis)[kent], 'enters,', (wis+zis)[kdep], 'departs')
            self.tableau.update(kent, kdep)
            if kdep == 2*self.nbi:
                break
        print('==========')
        if kdep == 2*self.nbi:
            print('Solution found: z_0 departed basis.')
            zsol, _, _ = self.tableau.solution()
            wsol = self.M_i_j @ zsol[:-1] + self.q_i
            if verbose > 0:
                print('Complementarity: z.w = ' + str(zsol[:-1] @ wsol))
                for i in range(self.nbi): print('z_'+str(i+1)+' = ' + str(zsol[i]))
                for i in range(self.nbi): print('w_'+str(i+1)+' = ' + str(wsol[i]))
            return zsol[:-1], wsol
        elif counter == maxit:
            print('Solution not found: maximum number of iterations (' + str(maxit) + ') reached.')
            return None
        else:
            print('Solution not found: Ray termination.')
            return None


class Bimatrix_game:
    def __init__(self, A_i_j, B_i_j):
        if A_i_j.shape != B_i_j.shape:
            raise ValueError("A_i_j and B_i_j must be of the same size.")
        self.A_i_j = A_i_j
        self.B_i_j = B_i_j
        self.nbi,self.nbj = A_i_j.shape

    def create_penalty_game(zero_sum=False):
        penalty_data = np.array([[53.21, 71.35, 93.80],
                                 [90.26, 42.81, 86.12],
                                 [96.88, 100.0, 75.43]])
        if zero_sum:
            zerosum_game = Bimatrix_game(A_i_j = penalty_data, B_i_j = 100 - penalty_data)
            return zerosum_game
        else:
            nonzerosum_game = Bimatrix_game(A_i_j = penalty_data,
                                            B_i_j = np.array([[150, 100, 100],
                                                              [100, 150, 100],
                                                              [100, 100, 150]]) - penalty_data)
            return nonzerosum_game

    def is_NashEq(self, p_i, q_j, tol=1e-5):
        for i in range(self.nbi):
            if np.eye(self.nbi)[i] @ self.A_i_j @ q_j > p_i @ self.A_i_j @ q_j + tol:
                print('Pure strategy i =', i, 'beats p_i.')
                return False
        for j in range(self.nbj):
            if p_i @ self.B_i_j @ np.eye(self.nbj)[j] > p_i @ self.B_i_j @ q_j + tol:
                print('Pure strategy j =', j, 'beats q_j.')
                return False
        return True

    def zero_sum_solve(self, verbose=0):
        return Matrix_game(self.A_i_j).solve(verbose)

    def mangasarian_stone_solve(self, verbose=0):
        model=grb.Model()
        model.Params.OutputFlag = 0
        model.params.NonConvex = 2
        p_i = model.addMVar(shape = self.nbi)
        q_j = model.addMVar(shape = self.nbj)
        α = model.addMVar(shape = 1, lb = -grb.GRB.INFINITY)
        β = model.addMVar(shape = 1, lb = -grb.GRB.INFINITY)
        model.setObjective(α + β - p_i @ (self.A_i_j + self.B_i_j) @ q_j, sense = grb.GRB.MINIMIZE)
        model.addConstr(α * np.ones((self.nbi,1)) - self.A_i_j @ q_j >= 0)
        model.addConstr(β * np.ones((self.nbj,1)) - self.B_i_j.T @ p_i >= 0)
        model.addConstr(p_i.sum() == 1)
        model.addConstr(q_j.sum() == 1)
        model.optimize()
        sol = np.array(model.getAttr('x'))
        if verbose > 0: print('p_i =', sol[:self.nbi], '\nq_j =', sol[self.nbi:(self.nbi+self.nbj)])
        return {'p_i': sol[:self.nbi], 'q_j': sol[self.nbi:(self.nbi+self.nbj)],
                'val1': sol[-2], 'val2': sol[-1]}

    def lemke_howson_solve(self,verbose = 0):
        A_i_j = self.A_i_j - np.min(self.A_i_j) + 1     # ensures that matrices are positive
        B_i_j = self.B_i_j - np.min(self.B_i_j) + 1
        zks = ['x_' + str(i+1) for i in range(self.nbi)] + ['y_' + str(j+1) for j in range(self.nbj)]
        wks = ['r_' + str(i+1) for i in range(self.nbi)] + ['s_' + str(j+1) for j in range(self.nbj)]
        complements = list(len(zks)+np.arange(len(zks))) + list(np.arange(len(zks)))
        C_k_l = np.block([[np.zeros((self.nbi, self.nbi)), A_i_j],
                          [B_i_j.T, np.zeros((self.nbj, self.nbj))]])
        tab = Tableau(C_k_l, np.ones(self.nbi + self.nbj), np.zeros(self.nbi + self.nbj), wks, zks)
        kent = len(wks) # z_1 enters
        while True:
            kdep = tab.determine_departing(kent)
            if verbose > 1:
                print('Basis: ', [(wks+zks)[i] for i in tab.k_b])
                print((wks+zks)[kent], 'enters,', (wks+zks)[kdep], 'departs')
            tab.update(kent, kdep)
            if (complements[kent] not in tab.k_b) and (complements[kdep] in tab.k_b):
                break
            else:
                kent = complements[kdep]
        z_k, _, _ = tab.solution() # solution returns: x_j, y_i, x_j@self.c_j
        x_i, y_j = z_k[:self.nbi], z_k[self.nbi:]
        α = 1 / y_j.sum()
        β = 1 /  x_i.sum()
        p_i = x_i * β
        q_j = y_j * α
        return {'p_i': p_i, 'q_j': q_j,
                'val1': α + np.min(self.A_i_j) - 1,
                'val2': β + np.min(self.B_i_j) - 1}


class TwoBases:
    def __init__(self,Phi_z_a,M_z_a,q_z=None,remove_degeneracies=True,M=None,eps=1e-5):
        self.Phi_z_a,self.M_z_a = Phi_z_a,M_z_a
        if M is None:
            M = self.Phi_z_a.max()
        self.nbstep,self.M,self.eps = 1,M,eps
        self.nbz,self.nba = self.Phi_z_a.shape
        if q_z is  None:
            self.q_z = np.ones(self.nbz)
        else:
            self.q_z = q_z
        
        # remove degeneracies:
        if remove_degeneracies:
            self.Phi_z_a += np.arange(self.nba,0,-1)[None,:]* (self.Phi_z_a == self.M)
            self.q_z = self.q_z + np.arange(1,self.nbz+1)*self.eps
        # create an M and a Phi basis
        self.tableau_M = Tableau( self.M_z_a[:,self.nbz:self.nba], d_i = self.q_z )
        self.basis_Phi = list(range(self.nbz))
        ###
        
    def init_a_entering(self,a_removed):
        self.basis_Phi.remove(a_removed)
        a_entering = self.nbz+self.Phi_z_a[a_removed,self.nbz:].argmax()
        self.basis_Phi.append(a_entering)
        self.entvar = a_entering
        return a_entering
    
    def get_basis_M(self):
        return set(self.tableau_M.k_b)
    
    def get_basis_Phi(self):
        return set(self.basis_Phi)

    def is_standard_form(self):
        cond_1 = (np.diag(self.Phi_z_a)  == self.Phi_z_a.min(axis = 1) ).all() 
        cond_2 = ((self.Phi_z_a[:,:self.nbz] + np.diag([np.inf] * self.nbz)).min(axis=1) >= self.Phi_z_a[:,self.nbz:].max(axis=1)).all()
        return (cond_1 & cond_2)
        
    def p_z(self,basis=None):
        if basis is None:
            basis = self.get_basis_Phi()
        return self.Phi_z_a[:,list(basis)].min(axis = 1)    
    
    def musol_a(self,basis=None):
        if basis is None:
            basis = self.get_basis_M()
        B = self.M_z_a[:,list(basis)]
        mu_a = np.zeros(self.nba)
        mu_a[list(basis)] = np.linalg.solve(B,self.q_z)
        return mu_a

    def is_feasible_basis(self,basis):    
        try:
            if self.musol_a(list(basis) ).min()>=0:
                return True
        except np.linalg.LinAlgError:
            pass
        return False
        
    def is_ordinal_basis(self,basis):
        res, which =False,None
        if len(set(basis))==self.nbz:
            blocking = (self.Phi_z_a[:,basis].min(axis = 1)[:,None] < self.Phi_z_a).all(axis = 0)
            if blocking.any():
                which = np.where(blocking)
        return res, which

    def determine_entering(self,a_departing):
        self.nbstep += 1
        pbefore_z = self.p_z(self.basis_Phi)
        self.basis_Phi.remove(a_departing)
        pafter_z = self.p_z(self.basis_Phi)
        i0 = np.where(pbefore_z < pafter_z)[0][0]
        c0 = min([(c,self.Phi_z_a[i0,c]) for c in self.basis_Phi  ],key = lambda x: x[1])[0]
        zstar = [z for z in range(self.nbz) if pafter_z[z] == self.Phi_z_a[z,c0] and z != i0][0]
        eligible_columns = [c for c in range(self.nba) if min( [self.Phi_z_a[z,c] - pafter_z[z] for z in range(self.nbz) if z != zstar]) >0 ]
        a_entering = max([(c,self.Phi_z_a[zstar,c]) for c in eligible_columns], key = lambda x: x[1])[0]
        self.basis_Phi.append(a_entering)
        return a_entering

    def step(self,a_entering ,verbose= 0):
        a_departing = self.tableau_M.determine_departing(a_entering)
        self.tableau_M.update(a_entering,a_departing)
        
        if self.get_basis_M() ==self.get_basis_Phi():
            if verbose>0:
                print('Solution found in '+ str(self.nbstep)+' steps. Basis=',self.get_basis_Phi() )
            return False
            
        new_entcol = self.determine_entering(a_departing)

        if verbose>1:
            print('Step=', self.nbstep)
            print('M basis = ' ,self.get_basis_M() )
            print('Phi basis = ' ,self.get_basis_Phi() )
            print('p_z=',self.p_z(list(self.get_basis_Phi()) ))
            print('entering var (M)=',a_entering)
            print('departing var (M and Phi)=',a_departing)
            print('entering var (Phi)=',new_entcol)

        return new_entcol


    def solve(self,a_departing = 0, verbose=0):
        a_entering = self.init_a_entering(a_departing)
        while a_entering:
            a_entering = self.step(a_entering,verbose)
        return({'basis': self.get_basis_Phi(),
                'mu_a':self.musol_a(),
                'p_z':self.p_z()})



import numpy as np
import gurobipy as grb
import matplotlib.pyplot as plt


class Polyhedral():
    def __init__(self,y_t_k,v_t,ytilde_s_k= np.array([]),vtilde_s = np.array([]),namef = '',namesv = None,verbose=0):
        if namesv is None:
            namesv = 'x'
        self.nbt,self.nbk= y_t_k.shape
        self.namef = namef
        if type(namesv)==str:
            if self.nbk==1:
                self.namesv = [namesv[0]] 
            else:
                self.namesv = [namesv[0]+'_'+str(k) for k in range(self.nbk)] 
        elif (type(namesv)==list) & (len(namesv)==self.nbk):
            self.namesv = namesv
        else:
            raise Exception("Parameter namesv not provided under the right format.")
        if ytilde_s_k.shape == (0,):
            self.nbs = 0
            ytilde_s_k = ytilde_s_k.reshape((0,self.nbk))
        else:
            self.nbs = ytilde_s_k.shape[0]
        self.nbi = self.nbt+self.nbs
        self.nbj = self.nbi+self.nbk+1
        if self.nbk > self.nbi:
            print('Caution: dimension larger than number of constraints.')
        self.ytilde_s_k = ytilde_s_k
        self.y_t_k = y_t_k
        self.vtilde_s = vtilde_s.reshape(self.nbs)
        self.v_t = v_t.reshape(self.nbt)
        self.tableau_i_j = np.block([[self.y_t_k, -np.ones((self.nbt,1)), np.eye(self.nbt),np.zeros( (self.nbt,self.nbs) ) ],
                                     [self.ytilde_s_k, -np.zeros((self.nbs,1)), np.zeros( (self.nbs,self.nbt) ), np.eye(self.nbs) ]])
        self.rhs_i = np.concatenate([self.v_t,self.vtilde_s])
        j_n = list(range(self.nbk+1))
        m = grb.Model()
        m.setParam('OutputFlag', 0)
        x_j = m.addMVar(self.nbj, lb = (self.nbk+1)*[-grb.GRB.INFINITY]+self.nbi*[0])
        m.setObjective( x_j[:self.nbk]@(- self.y_t_k[0,:]) + x_j[self.nbk], sense = grb.GRB.MINIMIZE)
        m.addConstr(self.tableau_i_j @ x_j == self.rhs_i)
        m.optimize()
        if m.Status in (grb.GRB.INFEASIBLE,grb.GRB.INF_OR_UNBD):
            self.all_infinite = 1
            if verbose>0:
                print('Empty domain')
        else:
            self.all_infinite = 0
            self.j_n = [i for  (i,v) in enumerate(m.getVars() ) if v.vBasis == -1]
            if verbose>0:
                print('Initial nonbasic columns=',self.j_n)


    def val(self,x_k):
        if np.array(x_k).shape ==():
            x_k = np.array([x_k])
        if self.nbs > 0:
            if (self.ytilde_s_k @ x_k - self.vtilde_s).max()>0:
                return float('inf')
        return (self.y_t_k @ x_k - self.v_t).max()

    def grad(self,x_k):
        if np.array(x_k).shape ==():
            x_k = np.array([x_k])
        if self.nbs > 0:
            if (self.ytilde_s_k @ x_k - self.vtilde_s).max()>0:
                return float('inf')
        k = (self.y_t_k @ x_k - self.v_t).argmax()
        return self.y_t_k[:,k]

    def __repr__(self,num_digits = 2,with_name = True):
        if self.all_infinite == 1:
            return '+ infinity everywhere' 
        elif self.all_infinite == -1:
            return '- infinity everywhere' 
        else:
            from sympy import Symbol
            from mec.lp import round_expr
            x_k = [Symbol(namev) for namev in self.namesv]
            if self.nbt >1:
                obj = f'max{str({round_expr(e,num_digits) for e in list(self.y_t_k @ x_k - self.v_t)} )}'
            else:
                obj = str( (self.y_t_k @ x_k - self.v_t)[0])
            constrs = [f'{round_expr(self.ytilde_s_k[s,:] @ x_k,num_digits)} <= {round(self.vtilde_s[s],num_digits)}' for s in range(self.nbs)]
            vars_str = ''.join([ name+ (',' if i<self.nbk-1 else '') for (i,name) in enumerate(self.namesv) ])
            vars_str_with_parenthesis = ('(' if self.nbk>1 else '') + vars_str + (')' if self.nbk>1 else '') 
            if with_name:
                if self.namef == '':
                    result = 'Function: '
                else:
                    result = self.namef + '(' + vars_str + ')' 
            else:
                result = ''
            result += str(obj)
            if (self.nbk==1) and (self.nbs>0):
                result += '\n'+ ('Domain: '+ self.namesv[0]+f' in {self.domain1d(num_digits)}')
            elif (self.nbk>1) and (self.nbs>0):
                result += '\n'+ 'Domain: ' + vars_str_with_parenthesis +' s.t. ' +('\n' if self.nbs>1 else '')
                for c in constrs:
                    result +=  str(c)+ '\n'
            else: # if there are no constraints
                result += '\n'+ str('Domain: ' + vars_str_with_parenthesis + ' in R' + ('^'+str(self.nbk) if self.nbk>1 else '') +'.\n')
            return result

    def domain1d(self,num_digits = 2):
        xl = max([float('-inf')]+[self.vtilde_s[s] / self.ytilde_s_k[s] for s in range(self.nbs) if self.ytilde_s_k[s]<0])
        xu = min([float('inf')]+[self.vtilde_s[s] / self.ytilde_s_k[s] for s in range(self.nbs) if self.ytilde_s_k[s]>0])
        return(round(float(xl),2), round(float(xu),2))

    def plot1d(self, xl=-10,xu=10,verbose = 0):
        if self.nbk >1:
            print('not 1d.')
        xla = max([float('-inf')]+[self.vtilde_s[s] / self.ytilde_s_k[s] for s in range(self.nbs) if self.ytilde_s_k[s]<0])
        xua = min([float('inf')]+[self.vtilde_s[s] / self.ytilde_s_k[s] for s in range(self.nbs) if self.ytilde_s_k[s]>0])
        if verbose>0:
            print(f'Domain=({float(xla)},{float(xua)})')
        xl = max(xla,xl)
        xu = min(xua,xu)
        xs = np.linspace(xl, xu, 400)
        ys = [self.val(x) for x in xs]
        plt.plot(xs, ys, label=self.namef+f'({self.namesv[0]})')
        plt.xlabel(self.namesv[0])
        plt.ylabel(self.namef)
        plt.legend()
        plt.show()


    def j_b(self,j_n = None):
        if j_n is None:
            j_n = self.j_n
        return [j for j in range(self.nbj) if j not in j_n]


    def subtableau_i_b(self,j_n=None):
        return self.tableau_i_j[:,self.j_b(j_n) ]

    def subtableau_i_n(self, j_n=None):
        if j_n is None:
            j_n = self.j_n
        return self.tableau_i_j[:,j_n ]

    def basic_solution_i(self, j_n):
        return np.linalg.solve(self.subtableau_i_b(j_n),self.rhs_i)

    def basic_infinite_solution_i(self,j_n,jent):
        ient = jent - self.nbk - 1
        therhs_i = np.zeros(self.nbi)
        therhs_i[ient] = -1
        return np.linalg.solve(self.subtableau_i_b(j_n),therhs_i)


    def dictionary(self,j_n=None): # of the form x_b = sol_b - D_b_n @ x_n
        sol_b = np.linalg.solve(self.subtableau_i_b(j_n),self.rhs_i)
        D_b_n = np.linalg.solve(self.subtableau_i_b(j_n),self.subtableau_i_n(j_n))
        return (sol_b,D_b_n)

    def determine_departing(self,j_n,jent):
        nent = [n for (n,j) in enumerate(j_n) if j==jent][0]
        (sol_b,D_b_n) = self.dictionary(j_n)
        D_b = D_b_n[:,nent]
        thedic = {b: sol_b[b] / D_b[b]
                  for b in range(self.nbk+1,self.nbi) # the nbk+1 first basic variables are the x_k and u and are unconstrained 
                  if  D_b[b] >0}
        if len (thedic)==0:
            return -1
        else:
            bdep = min(thedic, key = thedic.get)
            return self.j_b(j_n)[bdep]

    def star(self, namesv = None):
        if self.all_infinite != 0:
            v = Polyhedral(np.zeros( (1,self.nbk) ) ,np.zeros(1),namesv = namesv )
            v.all_infinite = - self.all_infinite 
            return v
        import networkx as nx
        the_graph = nx.DiGraph()
        j_n =  self.j_n.copy()
        xu = self.basic_solution_i(j_n)[:(self.nbk+1)]
        the_dict = {frozenset(j_n): xu }
        labels_to_add = [j_n]
        the_graph.add_nodes_from([frozenset(j_n)] )
        while len(labels_to_add)>0:
            j_n = labels_to_add.pop()
            for jent in j_n:
                jdep = self.determine_departing(j_n,jent)
                jnext_n = list({jdep} | set(j_n) -  {jent})
                if jdep > -1: # the node jnext_n is central
                    if frozenset(jnext_n) not in the_dict.keys():
                        labels_to_add.append(jnext_n) # attach to labels_to_add
                        the_graph.add_edges_from([(frozenset(j_n),frozenset(jnext_n)) ]) # add to the graph
                        xu = self.basic_solution_i(jnext_n)[:(self.nbk+1)] # find info 
                        the_dict.update({frozenset(jnext_n): xu }) # add to the dictionary
                else: #jdep == -1 ; means the node is exterior
                    # do not attach to labels_to_add
                    the_graph.add_edges_from([(frozenset(j_n),frozenset(jnext_n)) ]) # add to the graph
                    xu = self.basic_infinite_solution_i(j_n,jent)[:(self.nbk+1)] # find info 
                    the_dict.update({frozenset(jnext_n): xu }) # add to the dictionary

        xtilde_list = []
        utilde_list = []
        x_list = []
        u_list = []
        for f in the_dict.keys():
            if -1 in f:
                (x_k,u) = (the_dict[f][:-1],the_dict[f][-1])
                if np.abs(x_k).sum()>0: # only attach if x_k is not the zero vector
                    xtilde_list.append(x_k)
                    utilde_list.append(u)
            else:
                x_list.append(the_dict[f][:-1])
                u_list.append(the_dict[f][-1])
        xtilde_m_k = np.array(xtilde_list)
        utilde_m = np.array(utilde_list)
        x_m_k =  np.array(x_list)
        u_m = np.array(u_list)
        
        if namesv is None:
            namesv = [ (chr(ord(name[0])+1)+name[1:]) for name in self.namesv]
        namefstar = (self.namef+'*' if len(self.namef) >0 else '')
        ustar = Polyhedral(x_m_k,u_m,xtilde_m_k,utilde_m,namef=namefstar,namesv=namesv )

        return (ustar)
        
        
    def __add__(self,u2):
        u1 = self
        if isinstance(u2,Polyhedral): # addition of two polyhedral functions
            if u1.nbk == u2.nbk:
                nbk = u1.nbk
            else:
                print('Dimensions do not match.')
            y_t1t2_k = (u1.y_t_k[:,None,:] + u2.y_t_k[None,:,:]).reshape((-1,nbk))
            v_t1t2 = (u1.y_t_k[:,None] + u2.y_t_k[None,:]).flatten()
            ytilde_s_k = np.block([[u1.ytilde_s_k],[u2.ytilde_s_k]])
            vtilde_s = np.concatenate([u1.vtilde_s,u2.vtilde_s])
            if (u1.namesv == u2.namesv):
                namesv = u1.namesv
            else:
                namesv = None
            usum = Polyhedral(y_t1t2_k, v_t1t2,ytilde_s_k,vtilde_s, namesv = namesv )
            return(usum)
            
        elif isinstance(u2,(int,float)): # addition of a polyhedral function and a scalar
            return Polyhedral(u1.y_t_k,u1.v_t - u2,u1.ytilde_s_k,u1.vtilde_s, namesv = u1.namesv)
        
        else:
            raise NotImplementedError("Can't add Polyhedral and {}".format(type(u2)))
        
    def __radd__(self,scal):
        return self.__add__(scal)
    
    def __sub__(self,scal):
        if isinstance(scal,(int,float)): 
            return Polyhedral(self.y_t_k,self.v_t + scal,self.ytilde_s_k,self.vtilde_s, namesv = self.namesv)
        else:
            raise NotImplementedError("Can't substract Polyhedral from {}".format(type(scal)))
    
    def __mul__(self,scal):
        if isinstance(scal,(int,float)): 
            return Polyhedral(scal* self.y_t_k,scal* self.v_t,self.ytilde_s_k,self.vtilde_s, namesv = self.namesv)
        else:
            raise NotImplementedError("Can't multiply Polyhedral and {}".format(type(scal)))
            
    def __rmul__(self,scal):
        return self.__mul__(scal)
        
    def __truediv__(self,scal):
        if isinstance(scal,(int,float)): 
            return Polyhedral( self.y_t_k / scal, self.v_t / scal,self.ytilde_s_k,self.vtilde_s, namesv = self.namesv)
        else:
            raise NotImplementedError("Can't divide {} by Polyhedral.".format(type(scal)))


    def __or__(self,u2): 
        u1 = self
        if isinstance(u2,Polyhedral): # maximum of two PHC functions
            if u1.nbk == u2.nbk:
                nbk = u1.nbk
            else:
                print('Dimensions do not match.')
            y_t_k = np.block([[u1.y_t_k],[u2.y_t_k]])
            v_t = np.concatenate([u1.v_t,u2.v_t])
            ytilde_s_k = np.block([[u1.ytilde_s_k],[u2.ytilde_s_k]])
            vtilde_s = np.concatenate([u1.vtilde_s,u2.vtilde_s])
            if (u1.namesv == u2.namesv):
                namesv = u1.namesv
            else:
                namesv = None
            umax = Polyhedral(y_t_k, v_t,ytilde_s_k,vtilde_s, namesv = namesv )
            return(umax)
            
        elif isinstance(u2,(int,float)): # addition of a polyhedral function and a scalar
            return u1.__or__(Polyhedral(np.zero((1,u1.nbk), np.array([-u2]) , namesv = u1.namesv )))
        
        else:
            raise NotImplementedError("Can't take the maximum of Polyhedral and {}".format(type(u2)))
        
    def __ror__(self,scal):
        return self.__or__(scal)


    def __xor__(self,u2): # inf-convolution of two PHC functions
        if isinstance(u2,Polyhedral):
            if self.nbk == u2.nbk:
                nbk = self.nbk
            else:
                print('Dimensions do not match.')
            return (self.star() + u2.star()).star()
            
        else:
            raise NotImplementedError("Can't take the inf-convolution of Polyhedral and {}".format(type(u2)))
    
    
    # def graph(self):
        # # build skeleton:
        # import networkx as nx
        # the_graph = nx.DiGraph()
        # j_n =  self.j_n.copy()
        # xu = self.basic_solution_i(j_n)[:(self.nbk+1)]
        # the_dict = {frozenset(j_n): xu }
        # labels_to_add = [j_n]
        # the_graph.add_nodes_from([frozenset(j_n)] )
        # while len(labels_to_add)>0:
            # j_n = labels_to_add.pop()
            # for jent in j_n:
                # jdep = self.determine_departing(j_n,jent)
                # jnext_n = list({jdep} | set(j_n) -  {jent})
                # if jdep > -1: # the node jnext_n is central
                    # if frozenset(jnext_n) not in the_dict.keys():
                        # labels_to_add.append(jnext_n) # attach to labels_to_add
                        # the_graph.add_edges_from([(frozenset(j_n),frozenset(jnext_n)) ]) # add to the graph
                        # xu = self.basic_solution_i(jnext_n)[:(self.nbk+1)] # find info 
                        # the_dict.update({frozenset(jnext_n): xu }) # add to the dictionary
                # else: #jdep == -1 ; means the node is exterior
                    # # do not attach to labels_to_add
                    # the_graph.add_edges_from([(frozenset(j_n),frozenset(jnext_n)) ]) # add to the graph
                    # xu = self.basic_infinite_solution_i(j_n,jent)[:(self.nbk+1)] # find info 
                    # the_dict.update({frozenset(jnext_n): xu }) # add to the dictionary
        
        # #### HERE: something to build the bipartite graph            


            
def polyhedral_from_strings(expr_fun_str ,expr_dom_strs = [], verbose= 0):
    # for example exammple: 
    # expr_fun_str = 'max(3*a+2*b-1, 4*a-b+3,7*a-3*b+9)'
    # expr_dom_strs = ['a+1 <= 0 ','b >= 1']
    import sympy
    expr_fun = sympy.sympify(expr_fun_str)
    expr_doms = [sympy.sympify(expr_dom_str) for expr_dom_str in expr_dom_strs]
    variables = sorted(expr_fun.free_symbols.union(*[expr_dom.free_symbols for expr_dom in expr_doms] ), key=lambda x: x.name)
    if verbose:
        print('Variables =' , variables)
    list_y = []
    list_v = []
    for expr in expr_fun.args:
        coeffs = expr.as_coefficients_dict() 
        list_y.append([float(coeffs.get(v,0)) for v in variables] )
        list_v.append(-float(coeffs.get(1,0)))
    y_t_k = np.array(list_y)
    v_t = np.array(list_v)
    if len(expr_dom_strs) == 0:
        return Polyhedral(y_t_k,v_t) 

    list_ytilde = []
    list_vtilde = []

    for expr in expr_doms:
        lhs, rhs = expr.args
        if (expr.func == sympy.core.relational.LessThan) or (expr.func == sympy.core.relational.StrictLessThan):
            diff = lhs - rhs
        elif (expr.func == sympy.core.relational.GreaterThan) or (expr.func == sympy.core.relational.StrictGreaterThan):
            diff = rhs - lhs
        else:
            print('Not expected format.')
        coeffs = diff.as_coefficients_dict()
        list_ytilde.append([float(coeffs.get(v,0)) for v in variables] )
        list_vtilde.append(-float(coeffs.get(1,0)))
    ytilde_s_k = np.array(list_ytilde)
    vtilde_s = np.array(list_vtilde)
    return Polyhedral(y_t_k,v_t,ytilde_s_k,vtilde_s,namesv = [str(var) for var in variables] )


    # def plot2d

    # def relational_graph
    
    # soft max regularization

    # Hessian and possibly higher order derivatives
    
    # power diagrams 



# Linear programming library

import pandas as pd
import numpy as np
import scipy.sparse as spr
import gurobipy as grb
import sympy
from sympy.solvers import solve
from sympy import *
import matplotlib.pyplot as plt
import tabulate as tb


def round_expr(expr, num_digits):
    return expr.xreplace({n : round(n, num_digits) for n in expr.atoms(Number)})

def limited_tabulate(data, headers=None, tablefmt='grid', max_rows=18, max_cols=14):
    if max_rows is not None and len(data) > max_rows:
        data = data[:max_rows]

    if max_cols is not None:
        if headers:
            headers = headers[:max_cols]
        data = [row[:max_cols] for row in data]

    return tb.tabulate(data, headers=headers, tablefmt=tablefmt)





class LP():
    def __init__(self, A_i_j, d_i, c_j = None, decision_var_names_j = None, slack_var_names_i = None):
        if c_j is None:
            c_j = np.zeros(A_i_j.shape[1])
        self.A_i_j = A_i_j
        self.nbi, self.nbj = A_i_j.shape
        self.nbk = self.nbi+self.nbj
        self.d_i = d_i
        self.c_j = c_j
        if decision_var_names_j is None:
            decision_var_names_j = ['x_'+str(j) for j in range(self.nbj)]
        if slack_var_names_i is None:
            slack_var_names_i = ['s_'+str(i) for i in range(self.nbi)]
        self.decision_var_names_j = decision_var_names_j
        self.slack_var_names_i = slack_var_names_i

    def gurobi_solve(self,verbose=0):
        m = grb.Model()
        if verbose == 0:
            m.setParam('OutputFlag', 0)
        xg_j = m.addMVar(self.nbj)
        m.setObjective(xg_j@self.c_j,sense=grb.GRB.MAXIMIZE)
        constr_i = m.addConstr(self.A_i_j @ xg_j <= self.d_i)
        m.optimize()
        return(xg_j.x,constr_i.pi,m.objVal)


    def plot2d (self, the_path=[], legend=True):
        if len(self.c_j) != 2:
            print('The number of variables differs from two.')
            return()
        x1max = min(di/self.A_i_j[i,0] for i, di in enumerate(self.d_i) if self.A_i_j[i,0] != 0 and di/self.A_i_j[i,0] >= 0)
        x2max = min(di/self.A_i_j[i,1] for i, di in enumerate(self.d_i) if self.A_i_j[i,1] != 0 and di/self.A_i_j[i,1] >= 0)
        x1, x2 = np.meshgrid(np.linspace(-.2*x1max, 1.4*x1max, 400), np.linspace(-.2*x2max, 1.4*x2max, 400))
        feasible_region = (x1 >= 0) & (x2 >= 0)
        for i, di in enumerate(self.d_i):
            feasible_region = feasible_region & (self.A_i_j[i,0] * x1 + self.A_i_j[i,1] * x2 <= di)
        fig, ax = plt.subplots(figsize=(5, 5))
        plt.contourf(x1, x2, np.where(feasible_region, self.c_j[0]*x1 + self.c_j[1]*x2, np.nan), 50, alpha = 0.5, cmap='gray_r', levels=30)
        for i, di in enumerate(self.d_i):
            if self.A_i_j[i,1] != 0:
                ax.plot(x1[0, :], di/self.A_i_j[i,1] - self.A_i_j[i,0]/self.A_i_j[i,1]*x1[0, :], label=self.slack_var_names_i[i]+' = 0')
            else:
                ax.axvline(di/self.A_i_j[i,0], label=self.slack_var_names_i[i]+' = 0')
        if the_path:
            ax.plot([a for (a,_) in the_path], [b for (_,b) in the_path], 'r--', label='Agorithm path')
            ax.scatter([a for (a,_) in the_path], [b for (_,b) in the_path], color='red')
        ax.set_xlim(-.2*x1max, 1.4*x1max), ax.set_ylim(-.2*x2max, 1.4*x2max)
        ax.set_xlabel(self.decision_var_names_j[0]), ax.set_ylabel(self.decision_var_names_j[1])
        ax.spines[ 'left' ].set_position('zero'), ax.spines['bottom'].set_position('zero')
        ax.spines['right'].set_color('none'), ax.spines['top'].set_color('none')
        if legend: ax.legend(loc='upper right')
        plt.show()



class Dictionary(LP):
    def __init__(self, A_i_j, d_i, c_j = None, slack_var_names_i=None,decision_var_names_j=None):
        # s_i = d_i - A_i_j @ x_j
        if d_i.min()<0:
            from warnings import warn
            warn('The array d_i has negative entries; zero is not a feasible solution.')
        LP.__init__(self, A_i_j, d_i, c_j, decision_var_names_j, slack_var_names_i)
        self.nonbasic = [Symbol(x) for x in self.decision_var_names_j]
        self.base = { Symbol('obj') : c_j @ self.nonbasic }
        slack_exprs_i = d_i  - A_i_j @ self.nonbasic
        self.base.update({Symbol(name): slack_exprs_i[i] for (i,name) in enumerate(self.slack_var_names_i) })

    def variables(self):
        return( list(self.base.keys())[1:] + self.nonbasic)

    def display(self):
        print('-------------------------- \nObjective and constraints:')
        for var in self.base:
            print(var, '=', round_expr(self.base[var],2))


    def primal_solution(self, verbose=0):
        x_j = np.zeros(self.nbj)
        for j,var in enumerate([Symbol(x) for x in self.decision_var_names_j]):
            x_j[j]=float( self.base.get(var,sympy.Integer(0)).subs([(variable,0) for variable in self.nonbasic]) )
            if verbose > 0:
                print(var, '=', x_j[j])
        return x_j

    def determine_entering(self):
        self.nonbasic.sort(key=str) # Bland's rule
        for entering_var in self.nonbasic:
            if diff(self.base[Symbol('obj')],entering_var) > 0 :
                return entering_var
        return None # If no entering variable found, None returned

    def determine_departing(self,entering_var):
        runmin = float('inf')
        departing_var = None
        for var in self.base.keys() - {Symbol('obj')}:
            the_expr_list = solve(self.base[var] - var,entering_var)
            if the_expr_list: # if one can invert the previous expression
                the_expr = the_expr_list[0] # express entering variable as a function of the other ones:
                val_entering_var = the_expr.subs([ (variable,0) for variable in [var]+self.nonbasic])
                if (val_entering_var >= 0) & (val_entering_var < runmin) :
                    runmin,departing_var = val_entering_var, var
        return departing_var # if no variable is found, None returned

    def pivot(self,entering_var,departing_var, verbose = 0):
        expr_entering = solve(self.base[departing_var] - departing_var,entering_var)[0]
        for var in self.base:
            self.base[var] = self.base[var].subs([(entering_var, expr_entering)])
        self.base[entering_var] = expr_entering
        del self.base[departing_var]
        self.nonbasic.remove(entering_var)
        self.nonbasic.append(departing_var)
        if verbose > 0:
            print('Entering = ' + str( entering_var)+'; departing = '+ str( departing_var))
        if verbose > 1:
            print(str( entering_var)+' = '+str(round_expr(expr_entering,2)))
        return expr_entering

    def step(self,verbose=0):
        entering_var = self.determine_entering()
        if entering_var is None:
            print('Optimal solution found.\n=======================')
            self.primal_solution(verbose)
        else:
            departing_var = self.determine_departing(entering_var)
            if departing_var is None:
                print('Unbounded solution.')
            else:
                expr_entering_var = self.pivot(entering_var,departing_var, verbose)
                return False # not finished
        return True # finished

    def dual_solution(self,verbose = 0):
        y_i = np.zeros(self.nbi)
        for i,slackvar in enumerate(self.slack_var_names_i):
            y_i[i] = - diff(self.base[Symbol('obj')],slackvar)
            if verbose > 0 and y_i[i] != 0:
                print('pi_'+str(i)+'=', y_i[i])
        return y_i


    def simplex_loop(self,verbose = 0):
        if self.d_i.min()<0:
            from warnings import warn
            warn('The array d_i has negative entries; zero is not a feasible solution.')
        if verbose >2:
            [x1,x2] = [Symbol(x) for x in self.decision_var_names_j]
            the_path = [self.primal_solution()]
        finished = False
        while not finished:
            finished = self.step()
            if verbose>2:
                the_path.append(self.primal_solution())
        objVal = self.base[Symbol('obj')].subs([ (variable,0) for variable in self.nonbasic])
        if verbose>0:
            print('\nValue = ' + str(objVal))
        if verbose >2:
            self.plot2d(the_path, legend=False)
        return (self.primal_solution(),self.dual_solution(),objVal)


class Tableau(LP):
    def __init__(self, A_i_j, d_i, c_j = None, slack_var_names_i = None, decision_var_names_j = None): # A_i_j @ x_j + s_i = d_i
        LP.__init__(self, A_i_j, d_i, c_j, decision_var_names_j, slack_var_names_i)
        self.nbi,self.nbj = A_i_j.shape
        self.nbk = self.nbi + self.nbj
        if c_j is None:
            c_j = np.zeros(self.nbj)
        if decision_var_names_j is None:
            decision_var_names_j = ['x_'+str(j) for j in range(self.nbj)]
        if slack_var_names_i is None:
            slack_var_names_i = ['s_'+str(i) for i in range(self.nbi)]
        self.names_all_variables =  self.slack_var_names_i + self.decision_var_names_j
        self.tableau = np.block( [[np.zeros((1,self.nbi)), c_j.reshape((1,-1)), 0],
                                  [np.eye(self.nbi), A_i_j, d_i.reshape((-1,1))]] )
        self.k_b = list(range(self.nbi)) # columns associated with basic variables
        self.i_b = list(range(1,1+self.nbi)) # rows associated with basic variables

    def display(self):
        tableau = []
        tableau.append( ['Obj'] + list(self.tableau[0,:]) )
        for b in range(self.nbi):
            tableau.append([self.names_all_variables[self.k_b[b]]]+list(self.tableau[self.i_b[b],:]) )
        print(limited_tabulate(tableau, headers=[''] + self.names_all_variables + ['RHS'], tablefmt="grid"))

    def determine_entering(self):
        for k in range(self.nbk):
            if self.tableau[0,k] > 0:
                return k
        return None # If no entering variable found, None returned

    #def determine_departing(self, kent): # Alfred
    #    thedic = {self.k_b[b]: self.tableau[self.i_b[b],-1] / self.tableau[self.i_b[b],kent]
    #              for b in range(self.nbi) if self.tableau[self.i_b[b],kent]>0}
    #    kdep = min(thedic, key = thedic.get)
    #    return kdep

    def determine_departing(self, kent):
        runmin, kdep = float('inf'), None
        for b in range(self.nbi):
            if self.tableau[self.i_b[b],kent] > 0:
                ratio = self.tableau[self.i_b[b],-1] / self.tableau[self.i_b[b],kent]
                if (ratio < runmin):
                    runmin, kdep = ratio, self.k_b[b]
        return kdep

    def update(self, kent, kdep):
        bdep = self.k_b.index(kdep)
        idep = self.i_b[bdep]
        self.tableau[idep,:] = self.tableau[idep,:] / self.tableau[idep,kent]
        for i in range(1+self.nbi):
            if i != idep:
                self.tableau[i,:]= self.tableau[i,:] - self.tableau[idep,:] * self.tableau[i,kent]
        self.k_b[bdep] = kent
        self.i_b[bdep] = idep

    def simplex_step(self,verbose=0):
        if verbose>1:
            self.display()
        kent = self.determine_entering()
        if kent is not None:
            kdep= self.determine_departing(kent)
            if verbose>0:
                bdep = int(np.where(self.k_b == kdep)[0])
                print('Entering=', self.names_all_variables[kent], 'Departing=',self.names_all_variables[self.i_b[bdep]],'Pivot=',(self.i_b[bdep],kent))
            self.update(kent,kdep)
        else:
            if verbose>0:
                print ('Optimal solution found.')
            if verbose>1:
                self.display()
        return (kent is not None) # returns false  if optimal solution; true otherwise

    def simplex_solve(self,verbose=0):
        if self.d_i.min()<0:
            from warnings import warn
            warn('The array d_i has negative entries; zero is not a feasible solution.')
        while self.simplex_step(verbose):
            pass
        return self.solution()

    def solution(self): # returns the primal variables x_j and dual variables y_i
        x_j, s_i = np.zeros(self.nbj), np.zeros(self.nbi)
        for b in range(self.nbi):
            if self.k_b[b]<self.nbi:
                s_i[self.k_b[b]] = self.tableau[self.i_b[b],-1]
            else:
                x_j[self.k_b[b]-self.nbi] = self.tableau[self.i_b[b],-1]
        y_i = - self.tableau[0,:self.nbi]
        return x_j, y_i, x_j@self.c_j

##########################################
######### Interior Point Methods #########
##########################################

class InteriorPoint():
    def __init__(self, A, b, c, current_point=None):
        self.A, self.b, self.c = A, b, c
        self.current_point = current_point
        self.α = 1 - (1/8)/(1/5 + np.sqrt(len(self.c))) # shrinkage coeff from Freund & Vera

    #    def strictly_feasible_solution(self):
    #        x = np.linalg.lstsq(self.A, self.b) # Ax < b
    #        s = .01*np.ones(len(self.c))
    #        y = np.linalg.lstsq(self.A.T, s + self.c) # A.T y > c
    #        return np.concatenate((x,y,s))

    def plot_path(self, the_path, legend=True):
        plot_path(self.A, self.b, self.c, the_path, legend)

    def update(self, verbose=0):
        x, y, s, θ = self.current_point
        Δy = np.linalg.solve(self.A @ np.diag(1/s) @ np.diag(x) @ self.A.T, θ * self.A @ (1/s) - self.b)
        Δs = self.A.T @ Δy
        Δx = - x - np.diag(1/s) @ np.diag(x) @ Δs + θ * (1/s)
        self.current_point = [x+Δx, y+Δy, s+Δs, self.α*θ]
        return self.current_point

    def IP_loop(self, tol=1e-6, verbose=0):
        current_point = self.current_point
        new_point = self.update()
        if all(abs(np.concatenate(new_point[:-1]) - np.concatenate(current_point[:-1])) < tol):
            print('Optimal solution found.\n=======================')
            if verbose > 0:
                for i in range(len(new_point[0])): print("x_" + str(i+1), "=", new_point[0][i])
            else:
                if verbose > 1:
                    for i in range(len(new_point[0])): print("x_" + str(i+1), "=", new_point[0][i])
                return False # not finished
        return True # finished


def two_phase(A_i_j,d_i, verbose = False):
    nbi,nbj = A_i_j.shape
    signs_i = np.minimum(2*np.sign(d_i)+1,1) # 1 if >=0, -1 else
    d_i = signs_i*d_i
    A_i_j = signs_i[:,None] * A_i_j
    the_tableau = Tableau(A_i_j, d_i, c_j = A_i_j.sum(axis= 0) )
    the_tableau.simplex_solve()
    if (min(the_tableau.k_b) >= nbi ):
        if verbose:
            print('Feasible.')
        return [k-nbi for k in the_tableau.k_b ]
    else:
        if verbose:
            print('Infeasible.')
        return None

    



import numpy as np
import scipy.sparse as spr
import gurobipy as grb
from time import time
from scipy.stats import entropy
from sklearn import linear_model
from scipy import optimize

def M_x_xy(nbx,nby): # added on D1
    return spr.kron(spr.identity(nbx), np.ones((1,nby)))

def M_y_xy(nbx,nby): # added on D1
    return spr.kron(np.ones((1,nbx)),spr.identity(nby))



def sum_xlogx(a): # added on D3
    s=a.sum()
    return s*np.log(s) - s * entropy(a.flatten(),axis=None)


def display( args ):
    μ_x_y,u_x,v_y,valobs,valtot,iterations, taken,name = args
    print('*'*60)
    print('*'* (30 -len(name) // 2) + '  '+name + '  ' + '*'*(26 - (1+len(name)) // 2 ) )
    print('*'*60)
    print('Converged in ', iterations, ' steps and ', taken, 's.')
    print('Sum(mu*Phi)+σ*Sum(mu*log(mu))= ', valtot)
    print('Sum(mu*Phi)                = ', valobs)
    print('*'*60)
    return 

def ObjFunc(uvlambda,scale,mkt): # added on D5
    nbx,nby,nbk=mkt.nbx,mkt.nby,mkt.nbk
    u_x,v_y,λ_k = uvlambda[0:nbx] / scale,uvlambda[nbx:(nbx+nby)]/scale,uvlambda[(nbx+nby):(nbx+nby+nbk)] / scale
    
    Phi_a = mkt.φ_a_k @ λ_k
    mu_xy = np.exp((Phi_a[0:(nbx*nby)].reshape((nbx,nby)) - u_x.reshape((-1,1))-v_y.reshape((1,-1)))/2).flatten()
    mu_x0 = np.exp(Phi_a[(nbx*nby):(nbx*nby+nbx)]-u_x)
    mu_0y = np.exp(Phi_a[-nby:]-v_y)
    
    return (mkt.n_x.dot(u_x)+mkt.m_y.dot(v_y) - (Phi_a * mkt.μhat_a).sum() + 2 * mu_xy.sum() + mu_x0.sum() + mu_0y.sum())


def grad_ObjFunc(uvlambda,scale,mkt): # added on D5
    nbx,nby,nbk=mkt.nbx,mkt.nby,mkt.nbk
    u_x,v_y,λ_k = uvlambda[0:nbx]/scale,uvlambda[nbx:(nbx+nby)]/scale,uvlambda[(nbx+nby):(nbx+nby+nbk)]/scale
    
    Phi_a = mkt.φ_a_k @ λ_k
    mu_x_y = np.exp((Phi_a[0:(nbx*nby)].reshape((nbx,nby)) - u_x.reshape((-1,1))-v_y.reshape((1,-1)))/2).reshape((nbx,nby))
    mu_x0 = np.exp(Phi_a[(nbx*nby):(nbx*nby+nbx)]-u_x)
    mu_0y = np.exp(Phi_a[-nby:]-v_y)
    mu_a = np.concatenate((mu_x_y.flatten(),mu_x0,mu_0y))
    
    grad_u = mkt.n_x - mu_x_y.sum(axis = 1) - mu_x0
    grad_v = mkt.m_y - mu_x_y.sum(axis = 0) - mu_0y
    grad_lambda = (mu_a-mkt.μhat_a ) @ mkt.φ_a_k
    
    return np.concatenate((grad_u,grad_v,grad_lambda.flatten()))/scale



class OTProblem: 
    def __init__(self,Φ_x_y,n_x=None,m_y=None):
        self.Φ_a = Φ_x_y.flatten()
        self.nbx , self.nby = Φ_x_y.shape # number of types on each parts
        self.nbz = self.nbx + self.nby # total number of types 
        self.nba = self.nbx * self.nby # number of arcs
        if n_x is None: 
            self.n_x = np.ones(self.nbx)
        else:
            self.n_x = n_x
        if m_y is None:
            self.m_y = np.ones(self.nby)
        else:
            self.m_y = m_y


    def M_z_a(self): # added on D1
        return spr.vstack([M_x_xy(self.nbx,self.nby),M_y_xy(self.nbx,self.nby)])


    def q_z(self): # added on D1
        return np.concatenate([self.n_x,self.m_y])


    def solve_full_lp(self, OutputFlag= True): # added on D1
        m=grb.Model()
        μ_a = m.addMVar(self.nba)
        m.setObjective(self.Φ_a @ μ_a, grb.GRB.MAXIMIZE)
        m.addConstr(self.M_z_a() @ μ_a == self.q_z())
        m.setParam( 'OutputFlag', OutputFlag )
        m.optimize()
        if m.status == grb.GRB.Status.OPTIMAL:
            μopt_a = np.array(m.getAttr('x')).reshape(self.nbx,self.nby)
            popt_z = m.getAttr('pi')
            uopt_x, vopt_y = popt_z[:self.nbx],popt_z[self.nbx:]
        return μopt_a.reshape((self.nbx,-1)),uopt_x,vopt_y


    def solve_partial_lp(self, OutputFlag = True):
        m=grb.Model()
        μ_a = m.addMVar(self.nba)
        m.setObjective(self.Φ_a @ μ_a, grb.GRB.MAXIMIZE)
        m.addConstr(self.M_z_a() @ μ_a <= self.q_z())
        m.setParam( 'OutputFlag', OutputFlag )
        m.optimize()
        if m.status == grb.GRB.Status.OPTIMAL:
            μopt_a = np.array(m.getAttr('x')).reshape(self.nbx,self.nby)
            popt_z = np.array(m.getAttr('pi'))
            uopt_x, vopt_y = popt_z[:self.nbx],popt_z[self.nbx:]
        return (μopt_a.reshape((self.nbx,-1)),uopt_x,vopt_y)


    def solve_dual_partial_multiobj_lp(self, favor_Y = True):
        if favor_Y:
            welfare_weights_z = np.concatenate([np.ones(self.nbx),-np.ones(self.nby) ]) # welfare weights for each individuals
        else:
            welfare_weights_z = np.concatenate([-np.ones(self.nbx),np.ones(self.nby) ]) # welfare weights for each individuals
        m=grb.Model()
        p_z = m.addMVar(self.nbz)
        m.ModelSense = grb.GRB.MINIMIZE
        m.setParam('OutputFlag', 0) 
        q_z = self.q_z()
        m.setObjectiveN(q_z @ p_z, 0, priority = 2,weight = 1)
        m.setObjectiveN(welfare_weights_z @ p_z,1,priority = 1, weight = 1)
        m.addConstr(self.M_z_a().T @ p_z >= self.Φ_a)
        m.optimize()  
        if m.status == grb.GRB.Status.OPTIMAL:
            #μopt_a = np.array(m.getAttr('Pi')).reshape(self.nbx,self.nby)
            popt_z = np.array(m.getAttr('x'))
            uopt_x, vopt_y = popt_z[:self.nbx],popt_z[self.nbx:]
        return (uopt_x,vopt_y) # μopt_a.reshape((self.nbx,-1))


    def matrixIPFP(self,σ , tol = 1e-9, maxite = 1e+06 ): # added on D3
        ptm = time()
        ite = 0
        K_x_y = np.exp(self.Φ_a / σ).reshape(self.nbx,-1)
        B_y = np.ones(self.nby)
        error = tol + 1
        while error > tol and ite < maxite:
            A_x = self.n_x / (K_x_y @ B_y)
            KA_y = (A_x.T @ K_x_y)
            error = (abs(KA_y * B_y / self.m_y)-1).max()
            B_y = self.m_y / KA_y
            ite = ite + 1
            
        u_x,v_y = - σ * np.log(A_x),- σ * np.log(B_y)
        μ_x_y = K_x_y * A_x.reshape((-1,1)) * B_y.reshape((1,-1))
        valobs = self.Φ_a.dot(μ_x_y.flatten())
        valtot =  valobs - σ * sum_xlogx(μ_x_y)
        taken = time() - ptm
        if ite >= maxite:
            print('Maximum number of iteations reached in matrix IPFP.')    
        return μ_x_y,u_x,v_y,valobs,valtot,ite, taken, 'matrix IPFP'


    def logdomainIPFP_with_LSE_trick(self,σ , tol = 1e-9, maxite = 1e+06 ): # added on D3
        ptm = time()
        ite = 0
        Φ_x_y = self.Φ_a.reshape(self.nbx,-1)
        v_y = np.zeros(self.nby)
        λ_x,ζ_y = - σ * np.log(self.n_x), - σ * np.log(self.m_y)
        error = tol + 1
        while error > tol and ite < maxite:
            vstar_x = (Φ_x_y - v_y.reshape((1,-1))).max( axis = 1)
            u_x = λ_x + vstar_x + σ * np.log( (np.exp((Φ_x_y - vstar_x.reshape((-1,1)) - v_y.reshape((1,-1)))/σ)).sum( axis=1) )
            ustar_y = (Φ_x_y - u_x.reshape((-1,1)) ).max( axis = 0)
            KA_y = (np.exp((Φ_x_y -u_x.reshape((-1,1)) - ustar_y.reshape((1,-1)) ) / σ)).sum(axis=0)
            error = np.max(np.abs(KA_y * np.exp( (ustar_y-v_y) / σ) / self.m_y - 1))
            v_y = ζ_y + ustar_y+ σ * np.log(KA_y)
            ite = ite + 1
        μ_x_y =np.exp((Φ_x_y -u_x.reshape((-1,1)) - v_y.reshape((1,-1)))/σ )
        valobs = self.Φ_a.dot(μ_x_y.flatten())
        valtot =  valobs - σ * sum_xlogx(μ_x_y)
        taken = time() - ptm
        if ite >= maxite:
            print('Maximum number of iteations reached in log-domain IPFP with LSE trick.')
        return μ_x_y,u_x,v_y,valobs,valtot,ite, taken, 'log-domain IPFP with LSE trick'


    def solveGLM(self,σ , tol = 1e-9): # added on D3
        ptm = time()
        muhat_a = (self.n_x.reshape((self.nbx,-1)) @ self.m_y.reshape((-1,self.nby))).flatten() / self.n_x.sum()
        ot_as_glm = linear_model.PoissonRegressor(fit_intercept=False,tol=tol ,verbose=3,alpha=0)
        ot_as_glm.fit( - self.M_z_a().T, muhat_a * np.exp(-self.Φ_a / σ) , sample_weight = np.exp(self.Φ_a / σ))
        
        p = σ * ot_as_glm.coef_
        u_x,v_y  = p[:self.nbx] - p[0], p[self.nbx:]+p[0]
        μ_x_y =np.exp((self.Φ_a.reshape((self.nbx,-1)) -u_x.reshape((-1,1)) - v_y.reshape((1,-1)))/σ )
        valobs = self.Φ_a.dot(μ_x_y.flatten())
        valtot =  valobs - σ * sum_xlogx(μ_x_y)
        taken = time() - ptm
        return μ_x_y,u_x,v_y,valobs,valtot,None, taken, 'GLM'



class TUlogit: # added on D5
    def __init__(self,nbx,nby,φ_a_k , μhat_a ):
        self.nba, self.nbk = φ_a_k.shape
        self.nbx, self.nby  = nbx, nby
        self.φ_a_k, self.μhat_a = φ_a_k, μhat_a
        self.d_a = np.append(2*np.ones(nbx*nby),np.ones(nbx+nby))
        μ_x_y = self.μhat_a[0:(nbx*nby)].reshape((nbx,nby))
        self.n_x  = μ_x_y.sum(axis = 1) + self.μhat_a[(nbx*nby):(nbx*nby+nbx)]
        self.m_y  = μ_x_y.sum(axis = 0) + self.μhat_a[-nby:]


    def C(self): # added on D5
        nbx, nby, nbk = self.nbx, self.nby, self.nbk
        B = spr.vstack([spr.hstack([spr.kron(spr.identity(nbx), np.ones((nby,1))),spr.kron(np.ones((nbx,1)),spr.identity(nby))]),
                        spr.hstack([spr.identity(nbx) , spr.csr_matrix((nbx,nby))]),
                        spr.hstack([spr.csr_matrix((nby,nbx)),spr.identity(nby) ])])
        return spr.hstack([-B,self.φ_a_k])


    def fit_glm(self,tol = 1e-12,pretest = False): # added on D5
        clf = linear_model.PoissonRegressor(fit_intercept=False,tol=tol ,verbose=3,alpha=0)
        clf.fit( spr.diags(1/ self.d_a) @ self.C(), self.μhat_a,sample_weight = self.d_a)
        return clf.coef_[0:-self.nbk], clf.coef_[-self.nbk:]


    def fit_diy(self,scale=1,tol=1e-9): # added on D5
        uvl_0 = np.repeat(0,self.nbx+self.nby+self.nbk)
        outcome = optimize.minimize(ObjFunc,jac = grad_ObjFunc,args=(scale,self), method = 'CG', x0=uvl_0,options={'gtol':tol,'disp': True})
        return  outcome['x'][0:-self.nbk]/scale,outcome['x'][-self.nbk:]/scale



    def isCoercive(self):
        def poisson_coerciveness(X,y):
            n,m = X.shape
            b = X.T @ y
            m=grb.Model()
            u = m.addMVar(shape = 1)
            mu = m.addMVar(shape=n)
            m.setObjective(1 * u, grb.GRB.MAXIMIZE)
            m.addConstr(X.T @ mu == b)
            m.addConstr( mu - np.ones((n,1)) @ u >= 0)
            m.optimize()
            if m.status == grb.GRB.Status.OPTIMAL:
                solution = np.array(m.getAttr('x'))
                u = solution[0]
            else:
                u = np.NaN
            return(u)
    
        return poisson_coerciveness( self.C(), self.μhat_a)

    def assess(self,θ):
        C= self.C()
        return np.abs((np.exp(C @ θ) - self.μhat_a).reshape((1,-1)) @ C).max()


# Network flow problems library
import networkx as nx
import matplotlib.pyplot as plt
import numpy as np
import scipy.sparse as sp
from mec.lp import Tableau, two_phase
from anytree import Node, RenderTree

def create_connected_dag(num_nodes, num_edges, zero_node = 0, seed=777):
    np.random.seed(seed)
      
    cont = True
    while cont:
        G = nx.DiGraph()
        G.add_nodes_from(range(num_nodes))
        while len(G.edges) < num_edges:
            a, b = np.random.choice(G.nodes, 2,replace=False)
            if not nx.has_path(G, b, a):
                G.add_edge(a, b)

        if not nx.is_weakly_connected(G):
            components = list(nx.weakly_connected_components(G))
            for i in range(len(components) - 1):
                a = np.random.choice(components[i])
                b = np.random.choice(components[i+1])
                G.add_edge(a, b)

        node_list = [ 'z'+str(node) for node in G.nodes]
        arcs_list = [(node_list[i],node_list[j]) for i,j in G.edges]
        basis = list(np.random.choice(list( range(num_edges)), num_nodes-1 ) )
        mu_a = np.zeros(num_edges)
        mu_a[basis] = np.random.randint(0, num_edges, num_nodes - 1 )
        c_a = np.random.randint(0, num_edges, num_edges)

        q_z = np.array(nx.incidence_matrix(G, oriented=True).todense() @ mu_a)
        if q_z[0] <0:
            cont= False

    return (node_list,arcs_list,c_a,q_z,G)



def create_connected_bipartite(nbx, nby,  zero_node = 0, seed=777):
    num_nodes = nbx + nby
    num_edges = nbx * nby
    np.random.seed(seed)
    cont = True
    G = nx.DiGraph()
    G.add_nodes_from(['x'+str(i) for i in range(nbx)], bipartite= 0)
    G.add_nodes_from(['y'+str(j) for j in range(nby)], bipartite= 1)
    G.add_edges_from([('x'+str(i),'y'+str(j)) for j in range(nby)  for i in range(nbx)] )

    
    node_list = list(G.nodes)
    arcs_list = list(G.edges)
    basis = list(np.random.choice(list( range(num_edges)), num_nodes-1 ) )
    mu_a = np.zeros(num_edges)
    mu_a[basis] = np.random.randint(0, num_edges, num_nodes - 1 )+1
    c_a = np.random.randint(0, num_edges, num_edges)+1

    q_z = np.array(nx.incidence_matrix(G, oriented=True).todense() @ mu_a)
    #if q_z[0] <0:
    #    cont= False
    return (node_list,arcs_list,c_a,q_z,G)


class Network_problem:
    def __init__(self, nodesList, arcsList, c_a, q_z, active_basis=None, zero_node=0, pos=None, seed=777, verbose=0):
        self.zero_node = zero_node
        self.nbz = len(nodesList)
        self.nba = len(arcsList)
        self.nodesList = nodesList
        self.arcsList = arcsList
        self.c_a = c_a
        self.nodesDict = {node:node_ind for (node_ind,node) in enumerate(self.nodesList)}
        self.arcsDict = {arc:arc_ind for (arc_ind,arc) in enumerate(self.arcsList)}
        if verbose>1:
            print('Number of nodes = '+str(self.nbz)+'; number of arcs = '+str(self.nba)+'.')

        data = np.concatenate([-np.ones(self.nba),np.ones(self.nba)])
        arcsIndices = list(range(self.nba))
        arcsOrigins = [self.nodesDict[o] for o,d in self.arcsList]
        arcsDestinations = [self.nodesDict[d] for o,d in self.arcsList]
        
        znotzero = [i for i in range(self.nbz) if i != zero_node]
        self.q_z = q_z
        self.q0_z = q_z[znotzero]
        
        self.nabla_a_z = np.array(sp.csr_matrix((data, (arcsIndices+arcsIndices, arcsOrigins+arcsDestinations)), shape = (self.nba,self.nbz)).todense())
        self.nabla0_a_z = self.nabla_a_z[:,znotzero]
        
        active_basis = two_phase(self.nabla0_a_z.T,self.q0_z)
        assert len(active_basis) == (self.nbz - 1)
        
        self.digraph = nx.DiGraph()
        self.digraph.add_edges_from(arcsList)
        if pos is None:
            pos = nx.spring_layout(self.digraph, seed=seed)
        self.pos = pos

        arcsNames = [str(x)+str(y) for (x,y) in self.arcsList]

        self.tableau = Tableau(A_i_j = np.asarray(np.linalg.solve(self.B(active_basis),self.N(active_basis))),
                               d_i = self.q0_z,
                               c_j = self.gain_a(active_basis)[self.nonbasis(active_basis)],
                               slack_var_names_i = [arcsNames[i] for i in active_basis],
                               decision_var_names_j = [arcsNames[i] for i in self.nonbasis(active_basis)])
        the_arcs_indices = list(active_basis) + list(set(range(self.nba)) - set(active_basis))
        self.a_k = the_arcs_indices
        self.k_a = {the_arcs_indices[k]:k for k in range(self.nba)}

        
        
    
    def draw(self, p_z = None,mu_a = None, gain_a = None, entering_a = None, departing_a = None,figsize=(50, 30)):
        edge_labels = {e:('c='+f"{c:.0f}") for (e,c) in zip(self.arcsList,self.c_a)  }
        if mu_a is not None:
            for (i,e) in enumerate(self.arcsList):
                if i in self.basis(): 
                    edge_labels[e] += '\nμ='+f"{mu_a[i]:.0f}"

        if gain_a is not None:
            for (i,e) in enumerate(self.arcsList):
                if gain_a[i]!=self.c_a[i]: 
                    edge_labels[e] += '\ng='+f"{gain_a[i]:.0f}"

            
        nx.draw_networkx_edge_labels(self.digraph,self.pos,
                                    edge_labels=edge_labels,
                                    font_color='red')
                                    
        nx.draw(self.digraph,self.pos,
                edgelist=[self.arcsList[i] for i  in range(self.nba) if i not in self.basis() ],
                style='dotted', 
                with_labels=False)
                                              
        nx.draw_networkx_edges(self.digraph, self.pos, edgelist=[self.arcsList[i] for i in self.basis()], 
                               edge_color='blue') #connectionstyle='arc3,rad=0.3'
        
        labels = {z: f"q={self.q_z[i]:.0f}"+'\n'+z for i,z in enumerate(self.nodesList) }
        label_pos = {z: (position[0], position[1] ) for z, position in self.pos.items()}        
        if p_z is not None:
            p_z = np.concatenate([np.zeros(1),p_z])
            labels = {z: labels[z]+ f"\np={p_z[i]:.0f}" for i,z in enumerate(self.nodesList)} 
        
        nx.draw_networkx_labels(self.digraph, label_pos, labels,font_size=10,verticalalignment = 'center')
            
        if entering_a is not None:
            nx.draw_networkx_edges(self.digraph, self.pos, edgelist=[self.arcsList[entering_a]], 
                               edge_color='green', connectionstyle='arc3,rad=0.3')
            
        if departing_a is not None:
            nx.draw_networkx_edges(self.digraph, self.pos, edgelist=[self.arcsList[departing_a]], 
                                   style='dotted',
                                   edge_color='white')

        plt.figure(figsize=figsize)
        plt.show()

    # \nabla_0^\top = [B, N] where B corresponds to the basis and N to the non basis

    def basis(self, basis = None):
        if basis is None:
            basis = [ self.a_k[k] for k in self.tableau.k_b]
        return (basis)
    
    def nonbasis(self, basis = None):
        return ([i for i in range(self.nba) if i not in self.basis(basis)])
    
    
    def B(self, basis = None):
        return(self.nabla0_a_z[self.basis(basis),:].T) 
    
    def N(self, basis = None): 
        return(self.nabla0_a_z[self.nonbasis(basis),:].T)
        
    
    def musol_a(self,basis = None):
        mu_a = np.zeros(self.nba)
        mu_a[self.basis(basis)] = np.linalg.solve(self.B(),self.q0_z)
        return mu_a
        
    def p0sol_z(self,basis = None):
        return (np.linalg.solve(self.B(basis).T,self.c_a[self.basis(basis)]))
    
    def gain_a(self, basis = None):
        p_z = self.p0sol_z(basis)
        g_a = self.c_a.copy()
        g_a[self.nonbasis(basis)] = self.N(basis).T @ p_z 
        return g_a
    
    def cost_improvement_a(self, basis = None):
        return( self.gain_a() - self.c_a )
        
    def determine_entering_arcs(self,basis = None):
        return( np.where(self.cost_improvement_a(basis) > 0 )[0].tolist() )
    
    def tableau_update(self,entering_a,departing_a):
        self.tableau.update(self.k_a[entering_a],self.k_a[departing_a])
    
    def determine_departing_arc(self,entering_a, basis=None):
        entering_k = self.k_a[entering_a]
        departing_k = self.tableau.determine_departing(entering_k)
        if departing_k is None:
            departing_a = None
        else:
            departing_a = self.a_k[departing_k]
        return (departing_a)
    
    def iterate(self,  draw = False, verbose=0):
        entering_as = self.determine_entering_arcs()
        print('entering = ',entering_as)
        if not entering_as:
            if verbose>0:
                print('Optimal solution found.\n=======================')
            if draw:
                mu_a,p_z,g_a = self.musol_a(),self.p0sol_z(),self.gain_a()
                self.draw(p_z = p_z,mu_a=mu_a)
            return(0)
        else:
            entering_a=entering_as[0]
            departing_a = self.determine_departing_arc(entering_a)
            print('entering_a=', entering_a,'departing_a=', departing_a)
            if departing_a is None:
                if verbose>0:
                    print('Unbounded solution.')
                return(1)
            else:
                if verbose>1:
                    print('entering=',entering_a)
                    print('departing=',departing_a)
                if draw:
                    mu_a,p_z,g_a = self.musol_a(),self.p0sol_z(),self.gain_a()
                    self.draw(p_z = p_z,mu_a=mu_a, gain_a = g_a, entering_a = entering_a, departing_a = departing_a)
                    
                self.tableau_update(entering_a,departing_a)
                self.tableau
                return(2)


class EQF_problem:
    def __init__(self, nodesList, arcsList, galois_xy, q_z, active_basis=None, zero_node=0, pos=None, seed=777, verbose=0):

        self.zero_node = zero_node
        self.nbz = len(nodesList)
        self.nba = len(arcsList)
        self.nodesList = nodesList
        self.arcsList = arcsList
        self.nodesDict = {node:node_ind for (node_ind,node) in enumerate(self.nodesList)}
        self.arcsDict = {arc:arc_ind for (arc_ind,arc) in enumerate(self.arcsList)}
        if verbose>1:
            print('Number of nodes = '+str(self.nbz)+'; number of arcs = '+str(self.nba)+'.')

        data = np.concatenate([-np.ones(self.nba),np.ones(self.nba)])
        arcsIndices = list(range(self.nba))
        arcsOrigins = [self.nodesDict[o] for o,d in self.arcsList]
        arcsDestinations = [self.nodesDict[d] for o,d in self.arcsList]
        
        znotzero = [i for i in range(self.nbz) if i != zero_node]
        self.q_z = q_z
        self.q0_z = q_z[znotzero]
        
        self.nabla_a_z = np.array(sp.csr_matrix((data, (arcsIndices+arcsIndices, arcsOrigins+arcsDestinations)), shape = (self.nba,self.nbz)).todense())
        self.nabla0_a_z = self.nabla_a_z[:,znotzero]
        
        self.basis = two_phase(self.nabla0_a_z.T,self.q0_z)
        assert len(self.basis) == (self.nbz - 1)
        
        self.digraph = nx.DiGraph()
        self.digraph.add_edges_from(arcsList)
        if pos is None:
            pos = nx.spring_layout(self.digraph, seed=seed)
        self.pos = pos

        #arcsNames = [str(x)+str(y) for (x,y) in self.arcsList]
        self.galois_xy = galois_xy
        self.create_pricing_tree()
        
        

    def create_pricing_tree(self, verbose = False):
        the_graph = nx.DiGraph()
        the_graph.add_edges_from([self.arcsList[a] for a in self.basis])

        self.tree = {}

        def create_anytree(node, parent=None):
            if node not in self.tree:
                self.tree[node] = Node(name=node, parent=parent)
            else:
                self.tree[node].parent = parent
            for child in list(the_graph.neighbors(node)) + list(the_graph.predecessors(node)):
                if parent is None :
                    create_anytree(child, self.tree[node])
                elif child != parent.name:  # Prevent going back to the parent
                    create_anytree(child, self.tree[node])
                    
        create_anytree(self.nodesList[self.zero_node])
        if verbose:
            self.print_pricing_tree()

    def print_pricing_tree(self):
        for pre, fill, node in RenderTree(self.tree[self.nodesList[self.zero_node]]):
            print("%s%s" % (pre, node.name))
    
    def psol_z(self, current_price=0):
        nodename = self.nodesList[self.zero_node]
        self.set_prices_r(nodename,current_price)
        p_z = np.zeros(self.nbz)
        for (z,thename) in enumerate(self.nodesList):
            p_z[z] = self.tree[thename].price
        return(p_z)
        

    def set_prices_r(self,nodename, current_price=0):
        self.tree[nodename].price = current_price
        for child in self.tree[nodename].children:
            self.set_prices_r(child.name,self.galois_xy[(child.name,nodename)](current_price))


    def cut_pricing_tree(self,a_exiting): # returns root of second connected component
        x,y = self.arcsList[a_exiting]
        print (x,y)
        if (self.tree[y].parent == self.tree[x]):
            return y
        elif (self.tree[x].parent == self.tree[y]):
            return x
        else:
            print('Error in pricing tree during cut phase.')
    
    def paste_pricing_tree(self,a_entering,z_oldroot):
        x,y = self.arcsList[a_entering]

        if (self.tree[z_oldroot] in self.tree[y].ancestors):
            z_newroot , z_prec = y,x
        elif (self.tree[z_oldroot] in self.tree[x].ancestors):
            z_newroot , z_prec = x,y
        else:
            print('Error in pricing tree during paste phase.')
        
        z = z_newroot
        while (z_prec != z_oldroot):
            znext = self.tree[z].parent.name
            self.tree[z].parent = self.tree[z_prec]
            z_prec = z
            z = znext
    
    def iterate(self,  draw = False, verbose=0):
    
        p_z = self.psol_z()
        cost_improvement_a = np.array([self.galois_xy[(x,y)] (p_z[self.nodesDict[y]]) - p_z[self.nodesDict[x]] for (x,y) in self.arcsList])
        entering_as = np.where(cost_improvement_a )[0].tolist() 
        print('entering = ',entering_as)
        if not entering_as:
            if verbose>0:
                print('Optimal solution found.\n=======================')
            if draw:
                mu_a,p_z,g_a = self.musol_a(),self.p0sol_z(),self.gain_a()
                self.draw(p_z = p_z,mu_a=mu_a)
            return(0)
        else:
            entering_a=entering_as[0]
            departing_a = self.determine_departing_arc(entering_a)
            print('entering_a=', entering_a,'departing_a=', departing_a)
            if departing_a is None:
                if verbose>0:
                    print('Unbounded solution.')
                return(1)
            else:
                if verbose>1:
                    print('entering=',entering_a)
                    print('departing=',departing_a)
                if draw:
                    mu_a,p_z,g_a = self.musol_a(),self.p0sol_z(),self.gain_a()
                    self.draw(p_z = p_z,mu_a=mu_a, gain_a = g_a, entering_a = entering_a, departing_a = departing_a)
                    
                z_oldroot = self.cut_pricing_tree(departing_a)
                self.paste_pricing_tree(entering_a,z_oldroot)
                self.basis.remove(departing_a)
                self.basis.append(entering_a)
                return(2)

##################################################################
##################################################################
##################################################################
                
class Bipartite_EQF_problem:
    def __init__(self, n_x, m_y, galois_xy, label_galois_xy=None, seed=777, verbose=0):
        self.n_x,self.m_y = n_x,m_y
        self.nbx,self.nby = len(n_x),len(m_y)
        self.nbz = self.nbx + self.nby
        self.nba = self.nbx*self.nby
        self.galois_xy = galois_xy
        self.label_galois_xy = label_galois_xy
        
        self.digraph = nx.DiGraph()
        self.digraph.add_nodes_from(['x'+str(x) for x in range(self.nbx)], bipartite=0)
        self.digraph.add_nodes_from(['y'+str(y) for y in range(self.nby)], bipartite=1)
        self.digraph.add_edges_from([('x'+str(x),'y'+str(y)) for x in range(self.nbx) for y in range(self.nby)])
        bottom_nodes, top_nodes = nx.bipartite.sets(self.digraph)
        self.pos = {}
        self.pos.update((node, (1, index)) for index, node in enumerate(bottom_nodes))  # Set one side for one set
        self.pos.update((node, (2, index)) for index, node in enumerate(top_nodes))

        self.create_pricing_tree()
        self.create_p_z()
   
    def draw(self, draw_prices=False, mu_a=None, plot_galois=False, entering_a=None, departing_a=None, gain_a=None, figsize=(50, 30)):
        nx.draw(self.digraph, self.pos, with_labels=False)

        if entering_a is not None:
            nx.draw_networkx_edges(self.digraph, self.pos, edgelist=[entering_a], #edgelist=[self.arcsList[entering_a]],
                                   edge_color='green', connectionstyle='arc3,rad=0.3')

        if departing_a is not None:
            nx.draw_networkx_edges(self.digraph, self.pos, edgelist=[departing_a], #edgelist=[self.arcsList[departing_a]],
                                   style='dotted', edge_color='white')

        q_z = np.concatenate([self.n_x,self.m_y])
        labels = {z: f"{ ('n'*self.nbx+ 'm'*self.nby)[k]}={q_z[k]:.0f}"+'\n'+z+'\n' for k,z in enumerate(self.digraph.nodes()) }
        if draw_prices:
            labels = {z: labels[z] + f"p={self.p_z[k]:.0f}" for k,z in enumerate(self.digraph.nodes())}
        nx.draw_networkx_labels(self.digraph, self.pos, labels,font_size=10,verticalalignment = 'center')

        edge_labels = {e: '' for e in self.digraph.edges()}
        if plot_galois:
            for e in self.digraph.edges():
                edge_labels[e] += self.label_galois_xy[e]
        #if mu_a is not None:
        #    for (i,e) in enumerate(self.digraph.edges()):
        #        if i in self.basis():
        #            edge_labels[e] += '\nμ='+f"{mu_a[i]:.0f}"
        if gain_a is not None:
            for (i,e) in enumerate(self.digraph.edges()):
                if gain_a[i]!=self.c_a[i]: 
                    edge_labels[e] += '\ng='+f"{gain_a[i]:.0f}"
        nx.draw_networkx_edge_labels(self.digraph, self.pos, edge_labels=edge_labels, font_color='red', label_pos=.8)

        plt.figure(figsize=figsize)
        plt.show()

    def create_pricing_tree(self, display_tree=False):
        x,y=0,0
        res_x,res_y = self.n_x.copy(),self.m_y.copy()
        current_parent = 'x'+str(x)
        current_parent_node = Node(name = current_parent, parent = None)
        root_node = current_parent_node
        self.tree = {current_parent: root_node}
        current_parent_node.price = 0
        while (x<self.nbx) & (y<self.nby):
            current_child = ('y'+str(y) if current_parent[0]=='x' else 'x'+str(x))
            current_child_node = Node(name = current_child, parent = current_parent_node)
            current_child_node.price = self.galois_xy[(current_child,current_parent)](current_parent_node.price)
            self.tree[current_child] = current_child_node
            if res_x[x] <= res_y[y]:
                res_x[x], res_y[y] = 0, res_y[y]-res_x[x]
                if current_parent[0]=='x':
                    current_parent = 'y'+str(y)
                    current_parent_node = current_child_node
                x = x+1
            else:
                res_x[x], res_y[y] = res_x[x]-res_y[y], 0
                if current_parent[0]=='y':
                    current_parent = 'x'+str(x)
                    current_parent_node = current_child_node
                y = y+1
        if display_tree:
            self.display_tree()
        return root_node
    
    def display_tree(self):
        for pre, fill, node in RenderTree(self.tree['x0']):
            print("%s%s%s%s" % (pre, node.name,', p=' , node.price))
    
    def create_p_z(self, current_price=0):
        p_z = np.zeros(self.nbz)
        for z in range(self.nbz):
            p_z[z] = self.tree[list(self.digraph.nodes())[z]].price
        self.p_z = p_z
        return p_z

    def cost_improvement_a(self):
        return np.array([ self.galois_xy[(x,y)](self.p_z[self.nbx + int(y[1:])]) - self.p_z[int(x[1:])] for (x,y) in self.digraph.edges() ])

    def cut_pricing_tree(self, departing_a): # returns root of second connected component
        x,y = departing_a
        if (self.tree[y].parent == self.tree[x]):
            return y
        elif (self.tree[x].parent == self.tree[y]):
            return x
        else:
            print('Error in pricing tree during cut phase.')
    
    def paste_pricing_tree(self, entering_a, z_oldroot):
        x,y = entering_a
        if (self.tree[z_oldroot] in self.tree[y].ancestors):
            z_newroot, z_prec = y,x
        elif (self.tree[z_oldroot] in self.tree[x].ancestors):
            z_newroot, z_prec = x,y
        else:
            print('Error in pricing tree during paste phase.')
            return
        
        z = z_newroot
        while (z_prec != z_oldroot):
            znext = self.tree[z].parent.name
            self.tree[z].parent = self.tree[z_prec]
            z_prec = z
            z = znext

    def determine_departing_arc(self, entering_a, basis=None):
            entering_k = self.k_a[entering_a]
        departing_k = self.tableau.determine_departing(entering_k)
        if departing_k is None:
            departing_a = None
        else:
            departing_a = self.a_k[departing_k]
        return (departing_a)
        
    def iterate(self, draw=False, verbose=0):
        cost_improvement_a = self.cost_improvement_a()
        #entering_as = list(self.digraph.edges())[np.where(cost_improvement_a)]
        entering_as = [list(self.digraph.edges())[a] for a in range(self.nba) if cost_improvement_a[a] > 0]
        print('entering =', entering_as)
        if not entering_as:
            if verbose>0:
                print('Optimal solution found.\n=======================')
            #if draw:
            #    mu_a,p_z,g_a = self.musol_a(),self.p0sol_z(),self.gain_a()
            #    self.draw(p_z = p_z,mu_a=mu_a)
            return 0
        else:
            entering_a = entering_as[0]
            departing_a = self.determine_departing_arc(entering_a)
            print('entering_a =', entering_a,', departing_a =', departing_a)
            if departing_a is None:
                if verbose>0:
                    print('Unbounded solution.')
                return 1
            else:
                if verbose>1:
                    print('entering=',entering_a)
                    print('departing=',departing_a)
                if draw:
                    mu_a,p_z,g_a = self.musol_a(),self.p0sol_z(),self.gain_a()
                    self.draw(p_z = p_z,mu_a=mu_a, gain_a = g_a, entering_a = entering_a, departing_a = departing_a)
                    
                z_oldroot = self.cut_pricing_tree(departing_a)
                self.paste_pricing_tree(entering_a,z_oldroot)
                self.basis.remove(departing_a)
                self.basis.append(entering_a)
                return 2
                
                




import numpy as np
import networkx as nx
import scipy.optimize as opt
from time import time

def list_prefs(x,xletter,yletter,αo_y):
    nby = len(αo_y) - 1
    xsprefs = xletter + str(x) +' : ' 
    for y in range(nby):
        ind = np.where(αo_y== (nby - y ))[0][0]
        if (ind == nby):
            break        
        xsprefs = xsprefs + yletter + str( ind ) + ' > '
    return(xsprefs[:-3])

class NTU_market:
    def __init__(self,α_x_y,γ_x_y,n_x = np.array([]), m_y = np.array([])):
        nbx,nby = α_x_y.shape
        self.α_x_y = np.hstack((α_x_y,np.zeros((nbx,1) )))
        self.γ_x_y = np.vstack((γ_x_y,np.zeros( (1,nby) )))
        if n_x.size == 0:
            n_x = np.ones(nbx)
        if m_y.size ==  0:
            m_y = np.ones(nby) 
        self.n_x,self.m_y = n_x, m_y
        self.largex,self.largey = nby+1, nbx+1
        self.smallx,self.smally = -1, -1
        self.nbx,self.nby = nbx, nby
        self.αo_x_y = np.zeros((nbx,nby+1), dtype = 'int64') 
        self.γo_x_y = np.zeros((nbx+1,nby), dtype = 'int64') 
        self.prefslistα_x_y = np.zeros((nbx,nby+1), dtype = 'int64') 
        self.prefslistγ_x_y = np.zeros((nbx+1,nby), dtype = 'int64') 
        self.traceuo_x_t = np.array([])
        self.traceu_x_t = np.array([])
        for x in range(nbx):
            thelistx = (- self.α_x_y )[x,:].argsort()
            self.αo_x_y[x, thelistx] =  nby  - np.arange(nby+1)
            self.prefslistα_x_y[x,:] = (thelistx) % (self.nby+1) 
        for y in range(nby):
            thelisty = ( - self.γ_x_y)[:,y].argsort()
            self.γo_x_y[thelisty, y] = nbx  - np.arange(nbx+1)
            self.prefslistγ_x_y[:,y] = (thelisty) % (self.nbx+1) 
        self.comp_nbsteps = -1
        self.comp_time = -1
        self.eq_μ_x_y = np.array([])   
        self.eq_u_x = np.array([])
        self.eq_v_y = np.array([])
        

    def print_prefs(self,xs=[],ys=[]):
        if xs == [] and ys==[] :
            xs = range(self.nbx)
            ys = range(self.nby)
        if xs != [] :
            for x in xs :
                print(list_prefs(x,'x','y',self.αo_x_y[x,:]))
            
        if xs != [] and ys != [] :
            print('===')
            
        if ys != [] :
            for y in ys :
                print(list_prefs(y,'y','x',self.γo_x_y[:,y]))
               
           
    def is_GS_stable(self, μ_x_y = None, output=0 ):
        if (min(self.n_x)<1) or (max(self.n_x)>1) or (min(self.m_y)<1) or (max(self.m_y)>1):
            if output > 0 :
                print('n_x or m_y do not only contain ones.')
            return(False)
        if (μ_x_y is None):
            μ_x_y = self.eq_μ_x_y
        μext_x_y0 = np.hstack([μ_x_y, 1-np.sum(μ_x_y,axis = 1).reshape(-1,1) ])
        μext_x0_y = np.vstack([μ_x_y, 1-np.sum(μ_x_y,axis = 0).reshape(1,-1) ])
        if (np.logical_and(μext_x_y0 != 0  , μext_x_y0 != 1 )).any() or (np.logical_and(μext_x0_y != 0  , μext_x0_y != 1 )).any():
            if output > 0 :
                print('The μ is not feasible.')
            return(False)
        uo_x = np.sum(μext_x_y0 * self.αo_x_y, axis = 1)
        vo_y = np.sum(μext_x0_y * self.γo_x_y, axis = 0)
        for x in range(self.nbx):
            for y in range(self.nby):
                if (self.αo_x_y[x,y] > uo_x[x]) and (self.γo_x_y[x,y] > vo_y[y]):
                    if output > 0 :
                        print('The matching is not stable. One blocking pair is: x'+str(x)+',y'+str(y)+'.')
                    return(False)

        for x in range(self.nbx):
            if (self.αo_x_y[x,self.nby] > uo_x[x]):
                if output > 0 :
                    print('The matching is not stable. One blocking pair is: x'+str(x)+',y0.')
                return(False)

        for y in range(self.nby):
            if (self.γo_x_y[self.nbx,y] > vo_y[y]):
                if output > 0 :
                    print('The matching is not stable. One blocking pair is: x0,y'+str(y)+'.')
                return(False)
                
        if output > 0 :
            print('The matching is stable.')
        return (True)

    def μ_from_uo(self, uo_x):
        return np.where( (self.αo_x_y == uo_x.reshape((-1,1))) , 1 , 0 )[:,0:-1]

    def μ_from_vo(self, vo_y):
        return np.where( self.γo_x_y == vo_y , 1 , 0 )[0:-1,:]

    def uo_from_μ(self, μ_x_y = None):
        if (μ_x_y is None):
            μ_x_y = self.eq_μ_x_y
        return np.sum(np.hstack([μ_x_y, (self.n_x-np.sum(μ_x_y,axis = 1)).reshape(-1,1) ]) * self.αo_x_y, axis = 1)

    def vo_from_μ(self, μ_x_y = None):
        if (μ_x_y is None):
            μ_x_y = self.eq_μ_x_y
        return np.sum(np.vstack([μ_x_y, (self.m_y-np.sum(μ_x_y,axis = 0)).reshape(1,-1) ]) * self.γo_x_y, axis = 0)

    def u_from_μ(self, μ_x_y = None):
        if (μ_x_y is None):
            μ_x_y = self.eq_μ_x_y
        return np.sum(np.hstack([μ_x_y, (self.n_x-np.sum(μ_x_y,axis = 1)).reshape(-1,1) ]) * self.α_x_y, axis = 1)

    def v_from_μ(self, μ_x_y = None):
        if (μ_x_y is None):
            μ_x_y = self.eq_μ_x_y
        return np.sum(np.vstack([μ_x_y, (self.m_y-np.sum(μ_x_y,axis = 0)).reshape(1,-1) ]) * self.γ_x_y, axis = 0)

    def solveGaleShapley(self,output=0, trace=False):
        if (output>=2):
            print("Offers made and kept are denoted +1; offers made and rejected are denoted -1.")
        self.comp_nbsteps = 0
        tracemax = self.nbx*self.nby
        if trace:
            self.traceuo_x_t = np.zeros((tracemax,self.nbx))    
        μA_x_y = np.ones((self.nbx+1, self.nby+1), dtype = 'int64') # initially all offers are non rejected
        while True :
            μP_x_y = np.zeros((self.nbx+1, self.nby+1), dtype = 'int64')
            props_x = np.ma.masked_array(self.αo_x_y, μA_x_y[0:-1,:] ==0).argmax(axis = 1) # x's makes an offer to their favorite y
            μP_x_y[range(self.nbx),props_x] = 1 
            μP_x_y[self.nbx,0:self.nby] = 1
            
            μE_x_y = np.zeros((self.nbx+1, self.nby+1), dtype = 'int64') # y's retains their favorite offer among those made:
            rets_y = np.ma.masked_array(self.γo_x_y, μP_x_y[:,0:-1] ==0).argmax(axis = 0)
            μE_x_y[rets_y,range(self.nby)] = 1

            rej_x_y = μP_x_y - μE_x_y # compute rejected offers
            rej_x_y[self.nbx,:] = 0
            rej_x_y[:,self.nby] = 0

            μA_x_y = μA_x_y - rej_x_y  # offers from x that have been rejected are no longer available to x
            if output >= 2:
                print("Round "+str(self.comp_nbsteps)+":\n" , ( (2*μE_x_y-1) * μP_x_y)[0:-1,0:-1])
            if trace and self.comp_nbsteps<tracemax:
                for x in range(self.nbx):
                    self.traceuo_x_t[self.comp_nbsteps,x] = np.sum(self.αo_x_y [x,:] * μP_x_y[x,:])
            self.comp_nbsteps +=1
            if np.max(np.abs(rej_x_y)) == 0: 
                if trace:
                    self.traceuo_x_t = self.traceuo_x_t[0:min(self.comp_nbsteps,tracemax),:]
                break # if all offers have been accepted, then algorithm stops
        self.eq_μ_x_y = μE_x_y[0:-1,0:-1]
        return (0)

    def uo_from_vo(self,vo_y):
        excluded = np.hstack([(self.γo_x_y < vo_y)[0:-1,:],np.zeros((self.nbx,1))])
        return(np.ma.masked_array(self.αo_x_y,  excluded ).max(axis = 1).data)


    def  vo_from_uo(self,uo_x):
        excluded = np.vstack([(self.αo_x_y < uo_x.reshape((-1,1)))[:,0:-1],np.zeros((1,self.nby))])
        return(np.ma.masked_array(self.γo_x_y,  excluded ).max(axis = 0).data)
            
    def solveAdachi(self,output=0,trace=False, startu_x = None, startv_y = None ):
        self.comp_nbsteps = 0
        tracemax = self.nbx*self.nby
        if trace:
            self.traceuo_x_t = np.zeros((tracemax,self.nbx))    
        if startu_x is None:
            uo_x = self.largex * np.ones(self.nbx, dtype = 'int64') # x's utilities are highest
        else:
            uo_x = startu_x
        if startv_y is None:
            vo_y = self.smally * np.ones(self.nby, dtype = 'int64') # y's utilities are lowest
        else:
            vo_y = startv_y
        while True :
            uonew_x = self.uo_from_vo(vo_y) # each x proposes to favorite y among those willing to consider them
            if (uonew_x == uo_x).all() :
                if trace:
                    self.traceuo_x_t = self.traceuo_x_t[0:min(self.comp_nbsteps,tracemax),:]
                break
            uo_x = uonew_x      
            vo_y = self.vo_from_uo(uo_x) # each y proposes to favorite x among those willing to consider them
            if output >= 2:
                μP_x_y = self.μ_from_uo(uo_x)
                μE_x_y = self.μ_from_vo(vo_y)
                print("Round "+str(self.comp_nbsteps)+":\n","μ_P=\n" ,μP_x_y,"\n μ_E=\n", μE_x_y)
            if trace and self.comp_nbsteps<tracemax:
                self.traceuo_x_t[self.comp_nbsteps,:] = uo_x
            self.comp_nbsteps += 1
        self.eq_μ_x_y = self.μ_from_vo(vo_y)
        return(0)

    def Q_z(self,p_z):
        uo_x = p_z[0:self.nbx]
        vo_y = - p_z[self.nbx:(self.nbx+self.nby)]
        uonew_x = self.uo_from_vo(vo_y)
        vonew_y = self.vo_from_uo(uo_x)
        return(np.append(uo_x - uonew_x,vonew_y - vo_y))

    def cux_z(self,p_z):
        uo_x = p_z[0:self.nbx]
        vo_y = - p_z[self.nbx:(self.nbx+self.nby)]
        uonew_x = self.uo_from_vo(vo_y)
        return (np.append(uonew_x , - vo_y))

    def cuy_z(self,p_z):
        uo_x = p_z[0:self.nbx]
        vo_y = - p_z[self.nbx:(self.nbx+self.nby)]
        excluded = np.vstack([(self.αo_x_y < np.repeat([uo_x],self.nby+1,axis = 1).reshape((self.nbx,-1)))[:,0:-1],np.repeat(False,self.nby).reshape((-1,self.nby))])
        vonew_y = self.vo_from_uo(uo_x)
        return( np.append(uo_x, - vonew_y ) )

    def solveCU(self,output=0,trace=False):
        self.comp_nbsteps = 0
        tracemax = self.nbx*self.nby
        if trace:
            self.traceuo_x_t = np.zeros((tracemax,self.nbx))    
        p_z = np.append(self.largex * np.ones(self.nbx), - self.smally * np.ones(self.nby) )
        while True :
            pnew_z = self.cux_z(p_z) 
            pnew_z = self.cuy_z(pnew_z)
            if (pnew_z == p_z).all() :
                if trace:
                    self.traceuo_x_t = self.traceuo_x_t[0:min(self.comp_nbsteps,tracemax),:]
                break
            p_z = pnew_z      
            if trace and self.comp_nbsteps<tracemax:
                self.traceuo_x_t[self.comp_nbsteps,:] = p_z[0:self.nbx]
            self.comp_nbsteps += 1
        self.eq_μ_x_y = self.μ_from_vo(- p_z[self.nbx:(self.nbx+self.nby)])
        return(0)

    def damped_cux_z(self,p_z):
        uo_x = p_z[0:self.nbx]
        vo_y = - p_z[self.nbx:(self.nbx+self.nby)]
        uonew_x = self.uo_from_vo(vo_y)
        return (np.append(uo_x - np.where(uo_x > uonew_x, 1,0)+ np.where(uo_x < uonew_x, 1,0) , - vo_y))

    def solveDampedCU(self,output=0,trace=False):
        self.comp_nbsteps = 0
        tracemax = self.nbx*self.nby
        if trace:
            self.traceuo_x_t = np.zeros((tracemax,self.nbx))    
        p_z = np.append(self.largex * np.ones(self.nbx), - self.smally * np.ones(self.nby) )
        while True :
            pnew_z = self.damped_cux_z(p_z)   # each x proposes to favorite y among those willing to consider them:
            pnew_z = self.cuy_z(pnew_z)
            if (pnew_z == p_z).all() :
                if trace:
                    self.traceuo_x_t = self.traceuo_x_t[0:min(self.comp_nbsteps,tracemax),:]
                break
            p_z = pnew_z      
            # each y proposes to favorite x among those willing to consider them:
            if trace and self.comp_nbsteps<tracemax:
                self.traceuo_x_t[self.comp_nbsteps,:] = p_z[0:self.nbx]
            self.comp_nbsteps += 1
        self.eq_μ_x_y = self.μ_from_vo(- p_z[self.nbx:(self.nbx+self.nby)])
        return(0)


    def solveDeferredAcceptance(self, algorithm = 'Adachi', output=0, trace=False):
        start_time = time()
        if algorithm == 'GS' :
            self.solveGaleShapley(output,trace)
        elif algorithm == 'Adachi' :
            self.solveAdachi(output,trace)
        elif algorithm == 'DARUM' :
            self.solveDARUM(output,trace)   
        elif algorithm == 'CU' :
            self.solveCU(output,trace)
        else:
            raise Exception("Algorithm " + algorithm + " is not implemented.")
        self.comp_time =  time() - start_time
        if output >= 1:
            print("Converged in ",self.comp_nbsteps," steps and ", self.comp_time, " seconds.")
        if output == 1:
            print("mu_x_y=",self.eq_μ_x_y) 
        return (0)

       
    def matched_pairs(self, μ_x_y = None):
        if μ_x_y is None:
            μ_x_y = self.eq_μ_x_y
        nzx,nzy = np.nonzero(μ_x_y)
        return [(nzx[i],nzy[i]) for i in range(len(nzx) )]


    def next_us(self,uo_x,down = True):
        μ_x_y = self.μ_from_uo(uo_x)
        vo_y = self.vo_from_uo(uo_x)
        if down:
            incr= -1
        else:
            incr = 1
        g = nx.DiGraph()
        xs = list(range(self.nbx))
        ys = list(range(self.nbx,self.nby))
        g.add_nodes_from(xs+ys)
        g.add_edges_from([(x,self.nbx+y) for x,y in self.matched_pairs(μ_x_y)])
        #μprime_x_y = np.where( (self.αo_x_y.T == uo_x + incr).T , 1 , 0 )[:,0:-1]
        μprime_x_y = np.where( (self.αo_x_y[:,0:-1].T == uo_x + incr).T &  (self.γo_x_y[0:-1,:]*incr <= vo_y * incr), 1 , 0 ) 
        compatible_pairs = self.matched_pairs(μprime_x_y)
        g.add_edges_from([(self.nbx+j,i) for (i,j) in compatible_pairs])
        cycles = list(nx.simple_cycles(g))
        newus = []
        for c in cycles:
            xs_to_update = [x for x in c if (x <self.nbx)]
            nextu_x = uo_x.copy()
            nextu_x[xs_to_update] += incr
            newus.append(nextu_x)
        return(newus)

    def enumerate_us(self):
        self.solveGaleShapley()
        uo_x = np.sum(self.αo_x_y [:,0:-1] * self.eq_μ_x_y,axis = 1)
        us_list = [uo_x.tolist()]
        i = 0
        while (i < len(us_list)):
            inds_alreadyhere = []
            us_toadd = [arr.tolist() for arr in self.next_us(np.array(us_list[i],dtype='int64'))]
            for u_x in us_toadd:
                if u_x in us_list:
                    inds_alreadyhere.append(us_list.index(u_x))
                    us_toadd.remove(u_x)
            
            inds_toadd = list(range(len(us_list),len(us_list)+len(us_toadd) ))
            us_list = us_list + us_toadd 
            i += 1
        return(np.array([u for u in us_list],dtype = 'int64') )
        
    def DTF(self ,u_x,v_y):
        return np.maximum(u_x.reshape((-1,1)) - self.α_x_y[:,0:-1], v_y - self.γ_x_y[0:-1,:])
            
    def is_aggregate_stable(self, μ_x_y =None ,u_x=None,v_y =None, output=0 ):
        if μ_x_y is None:
            μ_x_y = self.eq_μ_x_y
        if u_x is None:
            u_x = self.eq_u_x
        if v_y is None:
            v_y = self.eq_v_y
        if (np.min(μ_x_y) < 0) or (np.min(self.n_x-np.sum(μ_x_y,axis = 1))<0) or (np.min(self.m_y-np.sum(μ_x_y,axis = 0))<0) :
            if output > 0 :
                print('The μ is not feasible.')
            return(False)
        if (np.min(u_x)<0) or  (np.min(v_y)<0):
            if output > 0 :
                print('u_x < 0 for some x or v_y < 0 for some y.')
            return(False)
        D_x_y = self.DTF(u_x,v_y)
        if np.min(D_x_y)<0 :
            if output > 0 :
                print('There is a blocking pair.')
            return(False)
        if np.sum(μ_x_y * D_x_y)>0 :
            if output > 0 :
                print('Complementary slackness does not hold.')
            return(False)
        if output > 0 :
            print('The matching is stable.')
        return (True)
        
    def aggregateChoice_noHet(self,axis,μbar_x_y):
        if axis == 0 : # if proposing side = x
            n_x = self.n_x
            nbx, nby = self.nbx, self.nby
            prefs_x_y,α_x_y = self.prefslistα_x_y, self.α_x_y
        else:
            n_x = self.m_y
            nbx, nby = self.nby, self.nbx
            prefs_x_y, α_x_y = self.prefslistγ_x_y.T, self.γ_x_y.T
        μ_x_y = np.zeros((nbx,nby+1))
        u_x = np.zeros(nbx)
        for x in range(nbx):
            nxres = n_x[x]
            for yind in prefs_x_y[x,]:
                if yind == nby:
                    μ_x_y[x,yind] = nxres
                    break
                if μbar_x_y[x,yind] > 0:
                    μ_x_y[x,yind] = min(nxres , μbar_x_y[x,yind])
                    nxres -= μ_x_y[x,yind]
                if nxres == 0:
                    break
            u_x[x]=α_x_y[x,yind]
        return(μ_x_y,u_x)

    def aggregateChoice_logit(self,axis,μbar_x_y):
        if axis == 0 : # if proposing side = x
            n_x = self.n_x
            nbx = self.nbx
            nby = self.nby
            α_x_y = self.α_x_y
        else:
            n_x = self.m_y
            nbx = self.nby
            nby = self.nbx
            α_x_y = self.γ_x_y.T    
        μ_x_y = np.zeros((nbx,nby+1))
        u_x = np.zeros(nbx)
        for x in range(nbx):
            thesolμ = opt.brentq(lambda theμ : (theμ+ np.minimum(theμ*np.exp(α_x_y[x,0:-1]),μbar_x_y[x,:]).sum() - n_x[x]) ,0,n_x[x])
            μ_x_y[x,0:-1] = np.minimum(thesolμ*np.exp(α_x_y[x,0:-1]),μbar_x_y[x,:])
            μ_x_y[x,nby] = thesolμ
            u_x[x] = - np.log(thesolμ / n_x[x])
        return(μ_x_y,u_x)

            
    def solveDARUM(self,het1 = 'none',het2 = 'none',output=0,trace=False,tol = 1e-5):
        self.comp_nbsteps = 0
        tracemax = 100*self.nbx*self.nby
        if (output>=2):
            print("Offers made and kept are denoted +1; offers made and rejected are denoted -1.")
        if trace:
            self.traceu_x_t = np.zeros((tracemax,self.nbx))
            self.traceuo_x_t = np.zeros((tracemax,self.nbx))    

        μA_x_y = np.array([[min(self.n_x[x],self.m_y[y]) for y in range(self.nby)] for x in range(self.nbx)]) 
        # initially all offers are non rejected
        while True :
            μP_x_y,self.eq_u_x = self.aggregateChoice(0 , μA_x_y,het1) # the x's pick their preferred offers among those not rejected
            μE_y_x,self.eq_v_y = self.aggregateChoice(1, μP_x_y[:,0:-1].T,het2)
            μE_x_y = μE_y_x.T # the y's pick their preferred offers among those made
            rej_x_y = μP_x_y[:,0:-1] - μE_x_y[0:-1,:] # compute rejected offers
            μA_x_y = μA_x_y - rej_x_y  # offers from x that have been rejected are no longer available to x
            if output >= 2:
                print("Round "+str(self.comp_nbsteps)+":\n" , ( (2*μE_x_y[0:-1,:]-1) * μP_x_y[:,0:-1]))            
            if trace and self.comp_nbsteps < tracemax:
                    self.traceu_x_t[self.comp_nbsteps,:] = self.eq_u_x
                    for x in range(self.nbx):
                        self.traceuo_x_t[self.comp_nbsteps,x] = np.sum(self.αo_x_y [x,:] * μP_x_y[x,:])
            self.comp_nbsteps +=1
            if np.max(np.abs(rej_x_y)) < tol: 
                if trace:
                    self.traceu_x_t = self.traceu_x_t[0:min(self.comp_nbsteps,tracemax),:]
                    self.traceuo_x_t = self.traceuo_x_t[0:min(self.comp_nbsteps,tracemax),:]
                break # if all offers have been accepted (within tolerange), then algorithm stops
        self.eq_μ_x_y = μE_x_y[0:-1,:]
        return (0)


    def aggregateChoice(self,axis,μbar_x_y,heterogeneity = 'none'):
        if heterogeneity == 'none':
            μ_x_y,u_x = self.aggregateChoice_noHet(axis,μbar_x_y)
        elif heterogeneity == 'logit':
            μ_x_y,u_x = self.aggregateChoice_logit(axis,μbar_x_y)
        else:
            raise Exception("Heterogeneity " + heterogeneity + " is not supported.")
        return (μ_x_y,u_x)


    def solveIPFP(self,output=0,trace=False,tol = 1e-5):
        self.comp_nbsteps = 0
        tracemax = 100*self.nbx*self.nby
        if trace:
            self.traceu_x_t = np.zeros((tracemax,self.nbx))
        μ_x0 = np.zeros(self.nbx)
        μ_0y = self.m_y.astype(np.float64)
        while True :
            for x in range(self.nbx):
                μ_x0[x] = opt.brentq (lambda theμ : (theμ+ np.minimum(theμ*np.exp(self.α_x_y[x,0:-1]),μ_0y*np.exp(self.γ_x_y[x,:])).sum() - self.n_x[x]) ,0,1.1*self.n_x[x])
            μP_x_y = np.minimum(μ_x0.reshape((-1,1))*np.exp(self.α_x_y[:,0:-1]),μ_0y*np.exp(self.γ_x_y[0:-1,:]))        
            self.eq_u_x = -np.log(μ_x0 / self.n_x)
            for y in range(self.nby):
                μ_0y[y] = opt.brentq(lambda theμ : (theμ+ np.minimum(μ_x0*np.exp(self.α_x_y[:,y]),theμ*np.exp(self.γ_x_y[0:-1,y])).sum() - self.m_y[y]) ,0,1.1*self.m_y[y])
            μE_x_y = np.minimum(μ_x0.reshape((-1,1))*np.exp(self.α_x_y[:,0:-1]),μ_0y*np.exp(self.γ_x_y[0:-1,:]))        
            self.eq_v_y = -np.log(μ_0y / self.m_y)
            rej_x_y = μP_x_y - μE_x_y # compute rejected offers
            if output >= 2:
                print('μP_x_y=\n',μP_x_y)            
                print('μE_x_y=\n',μE_x_y)            
            if trace and self.comp_nbsteps < tracemax:
                self.traceu_x_t[self.comp_nbsteps,:] = self.eq_u_x
            self.comp_nbsteps +=1
            if np.max(np.abs(rej_x_y)) < tol: 
                if trace:
                    self.traceu_x_t = self.traceu_x_t[0:min(self.comp_nbsteps,tracemax),:]
                break # if all offers have been accepted (within tolerance), then algorithm stops
        self.eq_μ_x_y = μE_x_y
        return (0)


class Aligned_NTU_market(NTU_market):
    def __init__(self,ϕ_x_y):
        NTU_market.__init__(self,α_x_y = ϕ_x_y, γ_x_y = ϕ_x_y)
        self.ϕ_x_y = ϕ_x_y
        self.n,_ =  ϕ_x_y.shape
        ϕo_x_y = np.zeros(self.n* self.n)
        ϕo_x_y[ϕ_x_y.flatten().argsort()] = 1+np.arange(self.n*self.n)
        self.ϕo_x_y=ϕo_x_y.reshape((self.n,self.n))
        
    def solveMaxMaxLex(self):
        theϕo_x_y = self.ϕo_x_y.copy()
        self.eq_μ_x_y = np.zeros((self.n,self.n), dtype = 'int64')
        for k in range(self.n):
            x,y = np.unravel_index(theϕo_x_y.argmax(),(self.n,self.n))
            self.eq_μ_x_y[x,y]=1
            theϕo_x_y[x,:]=0
            theϕo_x_y[:,y]=0
        return 0
        


import numpy as np
import pkg_resources
import pandas as pd
import tabulate as tb

def load_stigler_data(nbi = 9, nbj = 77, verbose=False):

    thepath =data_file_path = pkg_resources.resource_filename('mec', 'datasets/stigler-diet/StiglerData1939.txt')
    thedata = pd.read_csv(thepath , sep='\t')
    thedata = thedata.dropna(how = 'all')
    commodities = (thedata['Commodity'].values)[:-1]
    allowance = thedata.iloc[-1, 4:].fillna(0).transpose()
    nbi = min(len(allowance),nbi)
    nbj = min(len(commodities),nbj)
    if verbose:
        print('Daily nutrient content:')
        print(tb.tabulate(thedata.head()))
        print('\nDaily nutrient requirement:')
        print(allowance)
    return({'N_i_j':thedata.iloc[:nbj, 4:(4+nbi)].fillna(0).to_numpy().T,
            'd_i':np.array(allowance)[0:nbi],
            'c_j':np.ones(len(commodities))[0:nbj],
            'names_i': list(thedata.columns)[4:(4+nbi)],
            'names_j':commodities[0:nbj]}) 


def print_optimal_diet_stigler_data(q_j):
    print('***Optimal solution***')
    total,thelist = 0.0, []
    for j, commodity in enumerate(commodities):
        if q_j[j] > 0:
            total += q_j[j] * 365
            thelist.append([commodity,q_j[j]])
    thelist.append(['Total cost (optimal):', total])
    print(tb.tabulate(thelist))

def load_DupuyGalichon_data( verbose=False):
    thepath =data_file_path = pkg_resources.resource_filename('mec', 'datasets/marriage_personality-traits/')
    data_X = pd.read_csv(thepath + "Xvals.csv")
    data_Y = pd.read_csv(thepath + "Yvals.csv")
    aff_data = pd.read_csv(thepath + "affinitymatrix.csv")
    nbx,nbk = data_X.shape
    nby,nbl = data_Y.shape
    A_k_l = aff_data.iloc[0:nbk,1:nbl+1].values

    if verbose:
        print(data_X.head())
        print(data_Y.head())
        print(tb.tabulate(A_k_l))
        
    return({'data_X': data_X,
            'data_Y': data_Y,
            'A_k_l': A_k_l})
           
def load_ChooSiow_data(nbCateg = 25):
    thepath = pkg_resources.resource_filename('mec', 'datasets/marriage-ChooSiow/')
    n_singles = pd.read_csv(thepath+'n_singles.txt', sep='\t', header = None)
    marr = pd.read_csv(thepath+'marr.txt', sep='\t', header = None)
    navail = pd.read_csv(thepath+'n_avail.txt', sep='\t', header = None)
    μhat_x0 = np.array(n_singles[0].iloc[0:nbCateg])
    μhat_0y = np.array(n_singles[1].iloc[0:nbCateg])
    μhat_xy = np.array(marr.iloc[0:nbCateg:,0:nbCateg])
    Nhat = 2 * μhat_xy.sum() + μhat_x0.sum() + μhat_0y.sum()    
    μhat_a = np.concatenate([μhat_xy.flatten(),μhat_x0,μhat_0y]) / Nhat # rescale the data so that the total number of individual is one

    return({'μhat_a':μhat_a, 
             'Nhat':Nhat,
             'nbx':nbCateg,
             'nby':nbCateg
             }) 